import numpy as np
import pytest

from movis.attribute import AttributeType
from movis.motion import Motion
from movis import Easing


def test_motion_call():
    motion = Motion(init_value=0.5)
    assert np.all(motion(None, 0.0) == np.array([0.5]))
    assert np.all(motion(None, 1.0) == np.array([0.5]))
    motion = Motion(init_value=0.5, value_type=AttributeType.SCALAR)
    assert np.all(motion(None, 0.0) == np.array([0.5]))
    assert np.all(motion(None, 1.0) == np.array([0.5]))
    motion = Motion(init_value=0.5, value_type=AttributeType.ANGLE)
    assert np.all(motion(None, 0.0) == np.array([0.5]))
    assert np.all(motion(None, 1.0) == np.array([0.5]))
    motion = Motion(init_value=0.5, value_type=AttributeType.VECTOR2D)
    assert np.all(motion(None, 0.0) == np.array([0.5, 0.5]))
    motion = Motion(init_value=(0.5, 0.5), value_type=AttributeType.VECTOR2D)
    assert np.all(motion(None, 1.0) == np.array([0.5, 0.5]))
    motion = Motion(init_value=0.5, value_type=AttributeType.VECTOR3D)
    assert np.all(motion(None, 0.0) == np.array([0.5, 0.5, 0.5]))
    motion = Motion(init_value=(0.5, 0.5, 0.5), value_type=AttributeType.VECTOR3D)
    assert np.all(motion(None, 0.0) == np.array([0.5, 0.5, 0.5]))
    assert np.all(motion(None, 1.0) == np.array([0.5, 0.5, 0.5]))
    motion = Motion(init_value=0.5, value_type=AttributeType.COLOR)
    assert np.all(motion(None, 0.0) == np.array([0.5, 0.5, 0.5]))
    motion = Motion(init_value=(0.5, 0.5, 0.5), value_type=AttributeType.COLOR)
    assert np.all(motion(None, 0.0) == np.array([0.5, 0.5, 0.5]))
    assert np.all(motion(None, 1.0) == np.array([0.5, 0.5, 0.5]))


motion_append_params = [
    (AttributeType.SCALAR, 1),
    (AttributeType.ANGLE, 1),
    (AttributeType.VECTOR2D, 2),
    (AttributeType.VECTOR3D, 3),
    (AttributeType.COLOR, 3),
]


@pytest.mark.parametrize("value_type, n_dim", motion_append_params)
def test_motion_append(value_type, n_dim):
    motion = Motion(init_value=0.0, value_type=value_type)
    motion.append(0.0, 0.0)
    motion.append(1.0, 1.0)
    assert np.all(motion(None, -1.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.5) == np.array([0.5] * n_dim))
    assert np.all(motion(None, 1.0) == np.array([1.0] * n_dim))
    assert np.all(motion(None, 2.0) == np.array([1.0] * n_dim))


@pytest.mark.parametrize("value_type, n_dim", motion_append_params)
def test_motion_append_with_easing(value_type, n_dim):
    motion = Motion(init_value=0.0, value_type=value_type)
    motion.append(0.0, 0.0, easing=Easing.LINEAR)
    motion.append(1.0, 1.0, easing=Easing.LINEAR)
    assert np.all(motion(None, -1.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.5) == np.array([0.5] * n_dim))
    assert np.all(motion(None, 1.0) == np.array([1.0] * n_dim))
    assert np.all(motion(None, 2.0) == np.array([1.0] * n_dim))

    motion = Motion(init_value=0.0, value_type=value_type)
    motion.append(0.0, 0.0, easing='linear')
    motion.append(1.0, 1.0, easing='linear')
    assert np.all(motion(None, -1.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.5) == np.array([0.5] * n_dim))
    assert np.all(motion(None, 1.0) == np.array([1.0] * n_dim))
    assert np.all(motion(None, 2.0) == np.array([1.0] * n_dim))

    def easing_func(t):
        return t

    motion = Motion(init_value=0.0, value_type=value_type)
    motion.append(0.0, 0.0, easing=easing_func)
    motion.append(1.0, 1.0, easing=easing_func)
    assert np.all(motion(None, -1.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.5) == np.array([0.5] * n_dim))
    assert np.all(motion(None, 1.0) == np.array([1.0] * n_dim))
    assert np.all(motion(None, 2.0) == np.array([1.0] * n_dim))


@pytest.mark.parametrize("value_type, n_dim", motion_append_params)
def test_motion_append_full(value_type, n_dim):
    motion = Motion(init_value=[0.0] * n_dim, value_type=value_type)
    motion.append(0.0, [0.0] * n_dim)
    motion.append(1.0, [1.0] * n_dim)
    assert np.all(motion(None, -1.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.5) == np.array([0.5] * n_dim))
    assert np.all(motion(None, 1.0) == np.array([1.0] * n_dim))
    assert np.all(motion(None, 2.0) == np.array([1.0] * n_dim))


@pytest.mark.parametrize("value_type, n_dim", motion_append_params)
def test_motion_append_3(value_type, n_dim):
    motion = Motion(init_value=[0.0] * n_dim, value_type=value_type)
    motion.append(0.0, 0.0)
    motion.append(1.0, 1.0)
    assert np.all(motion(None, -1.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.5) == np.array([0.5] * n_dim))
    assert np.all(motion(None, 1.0) == np.array([1.0] * n_dim))
    assert np.all(motion(None, 2.0) == np.array([1.0] * n_dim))
    motion.append(0.5, 0.0)
    assert np.all(motion(None, -1.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.5) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.75) == np.array([0.5] * n_dim))
    assert np.all(motion(None, 1.0) == np.array([1.0] * n_dim))


@pytest.mark.parametrize("value_type, n_dim", motion_append_params)
def test_motion_extend(value_type, n_dim):
    motion = Motion(init_value=0.0, value_type=value_type)
    motion.extend(keyframes=[0.0, 1.0], values=[0.0, 1.0])
    assert np.all(motion(None, -1.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.5) == np.array([0.5] * n_dim))
    assert np.all(motion(None, 1.0) == np.array([1.0] * n_dim))
    assert np.all(motion(None, 2.0) == np.array([1.0] * n_dim))


@pytest.mark.parametrize("value_type, n_dim", motion_append_params)
def test_motion_extend_full(value_type, n_dim):
    motion = Motion(init_value=[0.0] * n_dim, value_type=value_type)
    motion.extend(
        keyframes=[0.0, 1.0],
        values=[[0.0] * n_dim, [1.0] * n_dim])
    assert np.all(motion(None, -1.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.5) == np.array([0.5] * n_dim))
    assert np.all(motion(None, 1.0) == np.array([1.0] * n_dim))
    assert np.all(motion(None, 2.0) == np.array([1.0] * n_dim))


@pytest.mark.parametrize("value_type, n_dim", motion_append_params)
def test_motion_extend_3(value_type, n_dim):
    motion = Motion(init_value=[0.0] * n_dim, value_type=value_type)
    motion.extend(keyframes=[0.0, 1.0], values=[0.0, 1.0])

    assert np.all(motion(None, -1.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.5) == np.array([0.5] * n_dim))
    assert np.all(motion(None, 1.0) == np.array([1.0] * n_dim))
    assert np.all(motion(None, 2.0) == np.array([1.0] * n_dim))

    motion.extend(keyframes=[0.5], values=[0.0])
    assert np.all(motion(None, -1.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.5) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.75) == np.array([0.5] * n_dim))
    assert np.all(motion(None, 1.0) == np.array([1.0] * n_dim))


@pytest.mark.parametrize("value_type, n_dim", motion_append_params)
def test_motion_clear(value_type, n_dim):
    motion = Motion(init_value=[0.0] * n_dim, value_type=value_type)
    motion.extend(keyframes=[0.0, 0.5, 1.0], values=[1.0, 1.0, 2.0])
    motion.clear()
    assert np.all(motion(None, -1.0) == np.array([0.0] * n_dim))


@pytest.mark.parametrize("value_type, n_dim", motion_append_params)
def test_motion_easing(value_type, n_dim):
    motion = Motion(init_value=[0.0] * n_dim, value_type=value_type)
    motion.extend(keyframes=[0.0, 1.0], values=[0.0, 1.0], easings=[Easing.LINEAR])

    assert np.all(motion(None, 0.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 1.0) == np.array([1.0] * n_dim))

    motion = Motion(init_value=[0.0] * n_dim, value_type=value_type)
    motion.extend(keyframes=[0.0, 1.0], values=[0.0, 1.0], easings=['linear'])

    assert np.all(motion(None, 0.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 1.0) == np.array([1.0] * n_dim))

    motion = Motion(init_value=[0.0] * n_dim, value_type=value_type)
    motion.extend(keyframes=[0.0, 1.0], values=[0.0, 1.0], easings=[Easing.EASE_IN_OUT])

    assert np.all(motion(None, 0.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.5) == np.array([0.5] * n_dim))
    assert np.all(motion(None, 1.0) == np.array([1.0] * n_dim))

    motion = Motion(init_value=[0.0] * n_dim, value_type=value_type)
    motion.extend(keyframes=[0.0, 1.0], values=[0.0, 1.0], easings=['ease_in_out'])

    assert np.all(motion(None, 0.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 0.5) == np.array([0.5] * n_dim))
    assert np.all(motion(None, 1.0) == np.array([1.0] * n_dim))

    motion = Motion(init_value=[0.0] * n_dim, value_type=value_type)
    motion.extend(keyframes=[0.0, 1.0], values=[0.0, 1.0], easings=[lambda t: t])

    assert np.all(motion(None, 0.0) == np.array([0.0] * n_dim))
    assert np.all(motion(None, 1.0) == np.array([1.0] * n_dim))



from __future__ import annotations

from typing import Callable, Hashable, Sequence

import numpy as np

from movis.enum import AttributeType
from movis.motion import Motion, transform_to_numpy


class Attribute:
    """Attribute class for animating the specified property.

    This class is used for animating layer properties.
    The dimensionality of the values that each Attribute can handle varies
    depending on the type of property.

    This is specified by setting an appropriate value for ``value_type`` using ``AttributeType`` Enum.
    For example, the values will be one-dimensional if ``value_type`` is set to ``AttributeType.SCALAR``.
    If set to ``AttributeType.COLOR``, the values will be three-dimensional.
    Regardless of ``value_type``, the returned type will always be ``numpy.ndarray``.

    .. note::
        Even if it's scalar, the returned value will be an array like ``np.array([value])``.

    Args:
        init_value:
            The initial value to use. It is used when no animation is set,
            or when an animation with zero keyframes is set.
        value_type:
            Specifies the type of value that the Attribute will handle.
            For more details, see the docs of ``AttributeType``.
        range:
            Defines the upper and lower limits of the possible values.
            If set, the returned array's values are guaranteed to be within this range;
            values that exceed this range are simply clipped.
        motion:
            The instance of the motion to use when adding keyframe animations.
            It can be specified in the constructor or activated later using ``enable_motion()``.
        functions:
            User-defined functions for adding animations, separate from keyframe animations.
            It can be specified in the constructor or added later using ``add_function()``.
    """
    def __init__(
        self,
        init_value: float | tuple[float, ...] | np.ndarray,
        value_type: AttributeType,
        range: tuple[float, float] | None = None,
        motion: Motion | None = None,
        functions: Sequence[Callable[[np.ndarray, float], np.ndarray]] | None = None,
    ) -> None:
        np_value = transform_to_numpy(init_value, value_type)
        clipped_value = np.clip(np_value, range[0], range[1]) if range is not None else np_value
        self._init_value: np.ndarray = clipped_value
        self._value_type = value_type
        self._range = range
        self._motion = motion
        self._functions = [] if functions is None else list(functions)

    def __call__(self, layer_time: float) -> np.ndarray:
        if self._motion is None and len(self._functions) == 0:
            return transform_to_numpy(self._init_value, self._value_type)
        else:
            value = self._init_value
            if self._motion is not None:
                value = transform_to_numpy(
                    self._motion(value, layer_time), self._value_type)
            if 0 < len(self._functions):
                for func in self._functions:
                    value = transform_to_numpy(
                        func(value, layer_time), self._value_type)
            if self._range is not None:
                return np.clip(value, self._range[0], self._range[1])
            else:
                return value

    def get_values(self, layer_times: np.ndarray) -> np.ndarray:
        """Returns an array of values for the specified layer times.

        Args:
            layer_times: An array of times for which to get the values.

        Returns:
            An array of values for the specified layer times.
        """
        assert layer_times.ndim == 1
        N, C = len(layer_times), len(self._init_value)
        if self._motion is None and len(self._functions) == 0:
            return np.broadcast_to(self._init_value.reshape(1, C), (N, C))
        else:
            values = [self(t) for t in layer_times]
            return np.concatenate(values).reshape(N, C)

    def enable_motion(self) -> Motion:
        """Enable `Motion` object to animate the attribute."""
        if self._motion is None:
            motion = Motion(init_value=self._init_value, value_type=self._value_type)
            self._motion = motion
        return self._motion

    def disable_motion(self) -> None:
        """Remove `Motion` object."""
        self._motion = None

    @property
    def init_value(self) -> np.ndarray:
        """The initial value of the attribute."""
        return self._init_value

    @init_value.setter
    def init_value(self, value: float | tuple[float, ...] | np.ndarray) -> None:
        value = transform_to_numpy(value, self._value_type)
        self._init_value = value
        if self._motion is not None:
            self._motion.init_value = value

    def set(self, init_value: float | Sequence[float] | np.ndarray) -> None:
        """Set the initial value of the attribute.

        .. note::
            This method is equivalent to ``init_value = value``.

        Args:
            init_value: The value to set.

        Examples:
            >>> import movis as mv
            >>> layer = mv.layer.Rectangle(size=(100, 100), color=(255, 0, 0))
            >>> layer.size.set((200, 200))
            >>> layer.color.set((0, 255, 0))
        """
        value = transform_to_numpy(init_value, self._value_type)
        self._init_value = value
        if self._motion is not None:
            self._motion.init_value = value

    @property
    def value_type(self) -> AttributeType:
        """The type of value that the attribute handles."""
        return self._value_type

    @property
    def range(self) -> tuple[float, float] | None:
        """The upper and lower limits of the possible values."""
        return self._range

    @range.setter
    def range(self, range: tuple[float, float] | None) -> None:
        self._range = range

    @property
    def motion(self) -> Motion | None:
        """The instance of the motion to use when adding keyframe animations."""
        return self._motion

    @property
    def functions(self) -> list[Callable[[np.ndarray, float], np.ndarray]]:
        """User-defined functions for adding animations, separate from keyframe animations."""
        return self._functions

    def add_function(
        self, function: Callable[[np.ndarray, float], np.ndarray]
    ) -> Callable[[np.ndarray, float], np.ndarray]:
        """Add a user-defined function for adding animations, separate from keyframe animations.

        Args:
            function: A function that takes two arguments, `value` and `time` and returns an array.

        Returns:
            The function that was added.
        """
        if not callable(function):
            raise ValueError(f"Invalid function: {function}")
        self._functions.append(function)
        return function

    def pop_function(self, index: int) -> Callable[[np.ndarray, float], np.ndarray]:
        """Remove a user-defined function of the specified index."""
        return self._functions.pop(index)

    def clear_functions(self) -> None:
        """Remove all user-defined functions."""
        self._functions.clear()

    def __getitem__(self, index: int) -> Callable[[np.ndarray, float], np.ndarray]:
        return self._functions[index]

    def __repr__(self) -> str:
        if self._motion is None:
            return f"{self._init_value}"
        else:
            return f"Attribute(value_type={self._value_type})"


class AttributesMixin:
    """A mix-in class with a collection of methods for implementing layers or effects using attributes.

    When using attributes, the content of layers or effects changes based on those attributes.
    Specifically, these attributes influence the generation of cache keys for the composition.

    This mixin class adds a feature that, when the ``get_key()`` method is called,
    extracts all attributes from the instance and converts them into a format that can be used for the cache key.
    """

    @property
    def attributes(self) -> dict[str, Attribute]:
        """A dictionary of attributes that are used to generate cache keys."""
        return {key: attr for key, attr in vars(self).items() if isinstance(attr, Attribute)}

    def get_key(self, time: float) -> tuple[Hashable, ...]:
        """Returns a tuple of hashable values that represent the state of the instance at a given time."""
        return tuple([transform_to_hashable(attr(time)) for attr in self.attributes.values()])


def transform_to_hashable(
    x: float | Sequence[float] | np.ndarray
) -> float | tuple[float, ...]:
    """Transform a scalar, tuple, list or numpy array to a hashable object used for caching.

    Args:
        x: The value to transform.

    Returns:
        A hashable object.
    """
    if isinstance(x, (int, float)):
        return float(x)
    elif len(x) == 1:
        return float(x[0])
    else:
        return tuple([float(v) for v in x])



from __future__ import annotations

import cv2
import numpy as np

from ..attribute import Attribute, AttributesMixin, AttributeType
from ..imgproc import alpha_composite
from ..util import to_rgb


class DropShadow(AttributesMixin):
    """Applies drop shadow to the input image.

    Args:
        radius:
            Radius of Gaussian kernel.
        offset:
            Offset of the shadow in pixels.
        angle:
            Angle of the shadow in degrees.
        color:
            Color of the shadow. It can be specified as a tuple of RGB values or a string of color name.
        opacity:
            Opacity of the shadow in the range ``[0, 1]``.

    Animatable Attributes:
        ``radius``
        ``offset``
        ``angle``
        ``color``
        ``opacity``
"""

    def __init__(
        self,
        radius: float = 0.0,
        offset: float = 0.0,
        angle: float = 45.0,
        color: tuple[int, int, int] | str = (0, 0, 0),
        opacity: float = 0.5
    ) -> None:
        self.radius = Attribute(radius, AttributeType.SCALAR, range=(0., 1e6))
        self.angle = Attribute(angle, AttributeType.SCALAR)
        self.offset = Attribute(offset, AttributeType.SCALAR)
        c = to_rgb(color)
        self.color = Attribute(c, AttributeType.COLOR, range=(0., 255.))
        self.opacity = Attribute(opacity, AttributeType.SCALAR, range=(0., 1.))

    def __call__(self, prev_image: np.ndarray, time: float) -> np.ndarray:
        assert prev_image.ndim == 3
        assert prev_image.shape[2] == 4, f'prev_image must be RGBA image, but {prev_image.shape}'

        radius = float(self.radius(time))
        assert 0 <= radius, f'radius must be nonnegative, but {radius}'

        # Compute ksize and create shadow image
        if 0. < radius:
            k = 2 * int(np.ceil(max(1, (4 * radius - 1) / 2))) + 1
            ksize = (k, k)
            alpha_shadow_image = cv2.GaussianBlur(
                src=np.pad(prev_image[:, :, 3], (ksize, ksize), mode='constant'),
                ksize=ksize, sigmaX=radius, sigmaY=radius)
        else:
            alpha_shadow_image = (self.opacity(time) * prev_image[:, :, 3]).astype(np.uint8)
        rgb_shadow_image = np.full(
            alpha_shadow_image.shape + (3,),
            np.round(self.color(time)).astype(np.uint8), dtype=np.uint8)
        alpha_shadow_image = (self.opacity(time) * alpha_shadow_image).astype(np.uint8)
        shadow_image = np.concatenate([rgb_shadow_image, alpha_shadow_image[:, :, None]], axis=2)

        # Compute offset and paste shadow image
        theta = float(self.angle(time)) * np.pi / 180
        vec = np.round(float(self.offset(time)) * np.array([np.cos(theta), np.sin(theta)])).astype(int)

        Hs, Ws = shadow_image.shape[:2]
        delta = np.abs(vec)
        new_image_shape = (Hs + 2 * delta[1], Ws + 2 * delta[0], 4)
        new_image = np.zeros(new_image_shape, dtype=np.uint8)
        p_shadow = delta + vec
        new_image[p_shadow[1]: (p_shadow[1] + Hs), p_shadow[0]: (p_shadow[0] + Ws)] = shadow_image

        # Put prev_image on the center of new_image
        p_image = (new_image_shape[1] - prev_image.shape[1]) // 2, (new_image_shape[0] - prev_image.shape[0]) // 2
        return alpha_composite(new_image, prev_image, p_image)



from __future__ import annotations

from os import PathLike
from typing import NamedTuple, Sequence

from .enum import Direction
from .util import to_rgb


class ASSStyleType(NamedTuple):
    name: str = 'Default'
    font_name: str = 'Helvetica'
    font_size: int = 60
    primary_color: str = '&Hffffff'
    secondary_color: str = '&Hffffff'
    outline_color: str = '&H0'
    back_color: str = '&H0'
    bold: bool = False
    italic: bool = False
    underline: bool = False
    strike_out: bool = False
    scale_x: int = 100
    scale_y: int = 100
    spacing: int = 0
    angle: int = 0
    border_style: int = 1
    outline: int = 5
    shadow: int = 0
    alignment: Direction = Direction.BOTTOM_CENTER
    margin_l: int = 10
    margin_r: int = 10
    margin_v: int = 30


def rgb_to_ass_color(color: str | tuple[int, int, int] | Sequence[int]) -> str:
    """Transform the rgb array into the string used in ASS file.

    Args:
        color:
            A union type that can be either a string or a tuple containing RGB integers or a sequence of integers.
            If ``str``: The string can be either a CSS color name (e.g., 'red') or a hexadecimal RGB string
            (e.g., ``'#FF0000'``). If ``tuple[int, int, int]`` or ``Sequence[int]``, it represents RGB values
            as integers between 0 and 255 (`e.g.`, ``(255, 0, 0)``).

    Returns:
        the color string used in ASS file.

    Example:
        >>> import movis as mv
        >>> mv.rgb_to_ass_color((255, 0, 0))
        '&H0000ff'
        >>> mv.rgb_to_ass_color('blue')
        '&Hff0000'
    """
    rgb_array = to_rgb(color)
    return "&H{:02x}{:02x}{:02x}".format(rgb_array[2], rgb_array[1], rgb_array[0])


def _make_ass_style_header():
    format_str = "Format: Name, Fontname, Fontsize, PrimaryColour, SecondaryColour, " \
        "OutlineColour, BackColour, Bold, Italic, Underline, StrikeOut, " \
        "ScaleX, ScaleY, Spacing, Angle, BorderStyle, " \
        "Outline, Shadow, Alignment, MarginL, MarginR, MarginV, Encoding"
    return '\n'.join(["[V4+ Styles]", format_str])


def _make_ass_style(s: ASSStyleType):
    style_str = f'Style: {s.name},{s.font_name},{s.font_size},{s.primary_color},{s.secondary_color},' \
        f'{s.outline_color},{s.back_color},' \
        f'{int(s.bold)},{int(s.italic)},{int(s.underline)},{int(s.strike_out)},' \
        f'{s.scale_x},{s.scale_y},{s.spacing},{s.angle},{s.border_style},' \
        f'{s.outline},{s.shadow},2,{s.margin_l},{s.margin_r},{s.margin_v},1'
    return style_str


def write_ass_file(
    start_times: Sequence[float],
    end_times: Sequence[float],
    texts: Sequence[str],
    dst_ass_file: str | PathLike,
    size: tuple[int, int] = (1920, 1080),
    characters: Sequence[str] | None = None,
    styles: Sequence[ASSStyleType] | None = None,
) -> None:
    """Writes an ASS (Advanced SubStation Alpha) subtitle file.

    Args:
        start_times:
            A list of start times in seconds for each subtitle entry.
        end_times:
            A list of end times in seconds for each subtitle entry.
        texts:
            A list of text strings corresponding to each subtitle entry.
        dst_ass_file:
            The destination path for the generated ASS file.
        size:
            The resolution of the video, defaults to (1920, 1080).
        characters:
            A list of character names for each subtitle entry. Defaults to None.
        styles:
            A list of styles for each subtitle entry. Defaults to None.

    Examples:
        >>> import movis as mv
        >>> mv.write_ass_file(
        ...     start_times=[0.0, 1.0, 3.0],
        ...     end_times=[1.0, 3.0, 4.0],
        ...     texts=['Subtitle1', 'Subtitle2', 'Subtitle3'],
        ...     styles=[mv.ASSStyleType(font_size=96)],  # Set default font size
        ...     dst_ass_file='subtitle1.ass')
        >>> mv.write_ass_file(
        ...    start_times=[0.0, 1.0, 3.0],
        ...    end_times=[1.0, 3.0, 4.0],
        ...    texts=['text by Alice', 'text by Bob', 'text by Alice'],
        ...    characters=['Alice', 'Bob', 'Alice'],
        ...    styles=[
        ...        mv.ASSStyleType(name='Alice', primary_color=mv.rgb_to_ass_color('red')),
        ...        mv.ASSStyleType(name='Bob', primary_color=mv.rgb_to_ass_color('blue'))],
        ...    dst_ass_file='subtitle2.ass')
    """
    assert len(start_times) == len(end_times) == len(texts)

    if characters is None:
        characters = ['Default'] * len(texts)

    if styles is None:
        styles = [ASSStyleType()]

    ass_style_header = _make_ass_style_header()
    ass_style_body = '\n'.join([_make_ass_style(style) for style in styles])

    header = f"""[Script Info]
; Script generated by FFmpeg/Lavc60.14.101
ScriptType: v4.00+
PlayResX: {size[0]}
PlayResY: {size[1]}
ScaledBorderAndShadow: yes
YCbCr Matrix: None

{ass_style_header}
{ass_style_body}

[Events]
Format: Layer, Start, End, Style, Name, MarginL, MarginR, MarginV, Effect, Text
"""
    line_template = "Dialogue: 0,{start_time},{end_time},{character},,0,0,0,,{text}"

    def get_time(t):
        return "{:02d}:{:02d}:{:02d}.{:02d}".format(
            int(t / 3600), int((t / 60) % 60), int(t % 60), int((t % 1) * 100)
        )

    lines = []
    for t0, t1, character, text in zip(start_times, end_times, characters, texts):
        text0, text1 = get_time(t0), get_time(t1)
        x = line_template.format(
            start_time=text0, end_time=text1, character=character, text=text)
        lines.append(x)
    body = "\n".join(lines)
    with open(dst_ass_file, "w") as fp:
        fp.write(header + body)


def write_srt_file(
        start_times: Sequence[float], end_times: Sequence[float], texts: Sequence[str],
        dst_srt_file: str | PathLike) -> None:
    """
    Writes an SRT (SubRip Text) subtitle file.

    Args:
        start_times:
            A list of start times in seconds for each subtitle entry.
        end_times:
            A list of end times in seconds for each subtitle entry.
        texts:
            A list of text strings corresponding to each subtitle entry.
        dst_srt_file:
            The destination path for the generated SRT file.
    """
    assert len(start_times) == len(end_times) == len(texts)
    with open(dst_srt_file, 'w') as srt:
        for i, (start_time, end_time, text) in enumerate(zip(start_times, end_times, texts)):
            srt.write('{}\n'.format(i + 1))
            srt.write('{:02d}:{:02d}:{:02d},{:03d} --> {:02d}:{:02d}:{:02d},{:03d}\n'.format(
                int(start_time / 3600), int((start_time / 60) % 60),
                int(start_time % 60), int((start_time % 1) * 1000),
                int(end_time / 3600), int((end_time / 60) % 60),
                int(end_time % 60), int((end_time % 1) * 1000),
            ))
            cleaned_text = text.replace(r"\n", "").replace("\n", "")
            srt.write(cleaned_text + '\n\n')



from enum import Enum


class CacheType(Enum):
    """A cache type used to determine how a result is cached during rendering."""
    COMPOSITION = 0
    LAYER = 1


class AttributeType(Enum):
    """A type used to determine the type and dimension of an attribute."""
    SCALAR = 0
    VECTOR2D = 1
    VECTOR3D = 2
    ANGLE = 3
    COLOR = 4

    @staticmethod
    def from_string(s: str) -> "AttributeType":
        """Convert a string to an attribute type."""
        if s == 'scalar':
            return AttributeType.SCALAR
        elif s == 'vector2d':
            return AttributeType.VECTOR2D
        elif s == 'vector3d':
            return AttributeType.VECTOR3D
        elif s == 'angle':
            return AttributeType.ANGLE
        else:
            raise ValueError(f"Unknown attribute type: {s}")


class Easing(Enum):
    """Constants for determining the completion function between keyframes.

    Constants:
        ``LINEAR``
            Linear interpolation between keyframes.
        ``EASE_IN``
            Ease in motion between keyframes. The larger the number, the more drastic the rate of change becomes.
        ``EASE_OUT``
            Ease out motion between keyframes. The larger the number, the more drastic the rate of change becomes.
        ``EASE_IN_OUT``
            Ease in and out motion between keyframes.
            The larger the number, the more drastic the rate of change becomes.
    """
    LINEAR = 0
    EASE_IN = 1
    EASE_OUT = 2
    EASE_IN_OUT = 3
    FLAT = 4
    EASE_IN2 = 102
    EASE_IN3 = 103
    EASE_IN4 = 104
    EASE_IN5 = 105
    EASE_IN6 = 106
    EASE_IN7 = 107
    EASE_IN8 = 108
    EASE_IN9 = 109
    EASE_IN10 = 110
    EASE_IN12 = 112
    EASE_IN14 = 114
    EASE_IN16 = 116
    EASE_IN18 = 118
    EASE_IN20 = 120
    EASE_IN25 = 125
    EASE_IN30 = 130
    EASE_IN35 = 135
    EASE_OUT2 = 202
    EASE_OUT3 = 203
    EASE_OUT4 = 204
    EASE_OUT5 = 205
    EASE_OUT6 = 206
    EASE_OUT7 = 207
    EASE_OUT8 = 208
    EASE_OUT9 = 209
    EASE_OUT10 = 210
    EASE_OUT12 = 212
    EASE_OUT14 = 214
    EASE_OUT16 = 216
    EASE_OUT18 = 218
    EASE_OUT20 = 220
    EASE_OUT25 = 225
    EASE_OUT30 = 230
    EASE_OUT35 = 235
    EASE_IN_OUT2 = 302
    EASE_IN_OUT3 = 303
    EASE_IN_OUT4 = 304
    EASE_IN_OUT5 = 305
    EASE_IN_OUT6 = 306
    EASE_IN_OUT7 = 307
    EASE_IN_OUT8 = 308
    EASE_IN_OUT9 = 309
    EASE_IN_OUT10 = 310
    EASE_IN_OUT12 = 312
    EASE_IN_OUT14 = 314
    EASE_IN_OUT16 = 316
    EASE_IN_OUT18 = 318
    EASE_IN_OUT20 = 320
    EASE_IN_OUT25 = 325
    EASE_IN_OUT30 = 330
    EASE_IN_OUT35 = 335

    @staticmethod
    def from_string(s: str) -> "Easing":
        """Convert a string to a motion type."""
        if s in STRING_TO_EASING:
            return STRING_TO_EASING[s]
        else:
            raise ValueError(f"Unknown motion type: {s}")


STRING_TO_EASING = {
    "linear": Easing.LINEAR,
    "ease_in": Easing.EASE_IN,
    "ease_out": Easing.EASE_OUT,
    "ease_in_out": Easing.EASE_IN_OUT,
    "flat": Easing.FLAT,
    "ease_in2": Easing.EASE_IN2,
    "ease_in3": Easing.EASE_IN3,
    "ease_in4": Easing.EASE_IN4,
    "ease_in5": Easing.EASE_IN5,
    "ease_in6": Easing.EASE_IN6,
    "ease_in7": Easing.EASE_IN7,
    "ease_in8": Easing.EASE_IN8,
    "ease_in9": Easing.EASE_IN9,
    "ease_in10": Easing.EASE_IN10,
    "ease_in12": Easing.EASE_IN12,
    "ease_in14": Easing.EASE_IN14,
    "ease_in16": Easing.EASE_IN16,
    "ease_in18": Easing.EASE_IN18,
    "ease_in20": Easing.EASE_IN20,
    "ease_in25": Easing.EASE_IN25,
    "ease_in30": Easing.EASE_IN30,
    "ease_in35": Easing.EASE_IN35,
    "ease_out2": Easing.EASE_OUT2,
    "ease_out3": Easing.EASE_OUT3,
    "ease_out4": Easing.EASE_OUT4,
    "ease_out5": Easing.EASE_OUT5,
    "ease_out6": Easing.EASE_OUT6,
    "ease_out7": Easing.EASE_OUT7,
    "ease_out8": Easing.EASE_OUT8,
    "ease_out9": Easing.EASE_OUT9,
    "ease_out10": Easing.EASE_OUT10,
    "ease_out12": Easing.EASE_OUT12,
    "ease_out14": Easing.EASE_OUT14,
    "ease_out16": Easing.EASE_OUT16,
    "ease_out18": Easing.EASE_OUT18,
    "ease_out20": Easing.EASE_OUT20,
    "ease_out25": Easing.EASE_OUT25,
    "ease_out30": Easing.EASE_OUT30,
    "ease_out35": Easing.EASE_OUT35,
    "ease_in_out2": Easing.EASE_IN_OUT2,
    "ease_in_out3": Easing.EASE_IN_OUT3,
    "ease_in_out4": Easing.EASE_IN_OUT4,
    "ease_in_out5": Easing.EASE_IN_OUT5,
    "ease_in_out6": Easing.EASE_IN_OUT6,
    "ease_in_out7": Easing.EASE_IN_OUT7,
    "ease_in_out8": Easing.EASE_IN_OUT8,
    "ease_in_out9": Easing.EASE_IN_OUT9,
    "ease_in_out10": Easing.EASE_IN_OUT10,
    "ease_in_out12": Easing.EASE_IN_OUT12,
    "ease_in_out14": Easing.EASE_IN_OUT14,
    "ease_in_out16": Easing.EASE_IN_OUT16,
    "ease_in_out18": Easing.EASE_IN_OUT18,
    "ease_in_out20": Easing.EASE_IN_OUT20,
    "ease_in_out25": Easing.EASE_IN_OUT25,
    "ease_in_out30": Easing.EASE_IN_OUT30,
    "ease_in_out35": Easing.EASE_IN_OUT35,
}


class BlendingMode(Enum):
    """Constants for determining the blending mode when compositing layers."""
    NORMAL = 0
    MULTIPLY = 1
    SCREEN = 2
    OVERLAY = 3
    DARKEN = 4
    LIGHTEN = 5
    COLOR_DODGE = 6
    COLOR_BURN = 7
    LINEAR_DODGE = 8
    LINEAR_BURN = 9
    HARD_LIGHT = 10
    SOFT_LIGHT = 11
    VIVID_LIGHT = 12
    LINEAR_LIGHT = 13
    PIN_LIGHT = 14
    DIFFERENCE = 15
    EXCLUSION = 16
    SUBTRACT = 17

    @staticmethod
    def from_string(s: str) -> "BlendingMode":
        """Convert a string to a blending mode."""
        if s in STRING_TO_BLENDING_MODE:
            return STRING_TO_BLENDING_MODE[s]
        else:
            raise ValueError(f"Unknown blending mode: {s}")


STRING_TO_BLENDING_MODE = {
    "normal": BlendingMode.NORMAL,
    "multiply": BlendingMode.MULTIPLY,
    "screen": BlendingMode.SCREEN,
    "overlay": BlendingMode.OVERLAY,
    "darken": BlendingMode.DARKEN,
    "lighten": BlendingMode.LIGHTEN,
    "color_dodge": BlendingMode.COLOR_DODGE,
    "color_burn": BlendingMode.COLOR_BURN,
    "linear_dodge": BlendingMode.LINEAR_DODGE,
    "linear_burn": BlendingMode.LINEAR_BURN,
    "hard_light": BlendingMode.HARD_LIGHT,
    "soft_light": BlendingMode.SOFT_LIGHT,
    "vivid_light": BlendingMode.VIVID_LIGHT,
    "linear_light": BlendingMode.LINEAR_LIGHT,
    "pin_light": BlendingMode.PIN_LIGHT,
    "difference": BlendingMode.DIFFERENCE,
    "exclusion": BlendingMode.EXCLUSION,
    "subtract": BlendingMode.SUBTRACT,
}


class MatteMode(Enum):
    """Constants for determining the matte mode when compositing layers."""
    NONE = 0
    ALPHA = 1
    LUMINANCE = 2

    @staticmethod
    def from_string(s: str) -> "MatteMode":
        if s in STRING_TO_MATTE_MODE:
            return STRING_TO_MATTE_MODE[s]
        else:
            raise ValueError(f"Unknown matte mode: {s}")


STRING_TO_MATTE_MODE = {
    "none": MatteMode.NONE,
    "alpha": MatteMode.ALPHA,
    "luminance": MatteMode.LUMINANCE,
}


class Direction(Enum):
    """Constants for determining the origin point of a layer."""
    BOTTOM_LEFT = 1
    BOTTOM_CENTER = 2
    BOTTOM_RIGHT = 3
    CENTER_LEFT = 4
    CENTER = 5
    CENTER_RIGHT = 6
    TOP_LEFT = 7
    TOP_CENTER = 8
    TOP_RIGHT = 9

    @staticmethod
    def from_string(s: str) -> "Direction":
        """Convert a string to a direction."""
        if s in STRING_TO_DIRECTION:
            return STRING_TO_DIRECTION[s]
        else:
            raise ValueError(f"Unknown origin point: {s}")

    @staticmethod
    def to_vector(d: "Direction", size: tuple[float, float]) -> tuple[float, float]:
        """Convert a direction to a vector."""
        if d == Direction.BOTTOM_LEFT:
            return (0, size[1])
        elif d == Direction.BOTTOM_CENTER:
            return (size[0] / 2, size[1])
        elif d == Direction.BOTTOM_RIGHT:
            return (size[0], size[1])
        elif d == Direction.CENTER_LEFT:
            return (0, size[1] / 2)
        elif d == Direction.CENTER:
            return (size[0] / 2, size[1] / 2)
        elif d == Direction.CENTER_RIGHT:
            return (size[0], size[1] / 2)
        elif d == Direction.TOP_LEFT:
            return (0, 0)
        elif d == Direction.TOP_CENTER:
            return (size[0] / 2, 0)
        elif d == Direction.TOP_RIGHT:
            return (size[0], 0)
        else:
            raise ValueError(f"Unknown direction: {d}")


STRING_TO_DIRECTION = {
    "bottom_left": Direction.BOTTOM_LEFT,
    "bottom_center": Direction.BOTTOM_CENTER,
    "bottom_right": Direction.BOTTOM_RIGHT,
    "center_left": Direction.CENTER_LEFT,
    "center": Direction.CENTER,
    "center_right": Direction.CENTER_RIGHT,
    "top_left": Direction.TOP_LEFT,
    "top_center": Direction.TOP_CENTER,
    "top_right": Direction.TOP_RIGHT,
}


class TextAlignment(Enum):
    """Constants for determining the alignment of text."""
    LEFT = 0
    CENTER = 1
    RIGHT = 2

    @staticmethod
    def from_string(s: str) -> "TextAlignment":
        """Convert a string to a text alignment."""
        if s in STRING_TO_TEXT_ALIGNMENT:
            return STRING_TO_TEXT_ALIGNMENT[s]
        else:
            raise ValueError(f"Unknown text alignment: {s}")


STRING_TO_TEXT_ALIGNMENT = {
    "left": TextAlignment.LEFT,
    "center": TextAlignment.CENTER,
    "right": TextAlignment.RIGHT,
}



from . import effect  # noqa
from . import layer  # noqa
from .attribute import (Attribute, AttributesMixin,  # noqa
                        transform_to_hashable)
from .enum import (AttributeType, BlendingMode, Direction, Easing,  # noqa
                   MatteMode, TextAlignment)
from .imgproc import alpha_composite  # noqa
from .layer.protocol import AUDIO_SAMPLING_RATE  # noqa
from .motion import Motion  # noqa
from .ops import concatenate, crop, insert, repeat, switch, tile, trim  # noqa
from .subtitle import (ASSStyleType, rgb_to_ass_color, write_ass_file,  # noqa
                       write_srt_file)
from .transform import Transform  # noqa
from .util import add_materials_to_video, to_rgb  # noqa



from __future__ import annotations

from collections.abc import Sequence as SequenceType
from os import PathLike
from typing import Sequence


def add_materials_to_video(
    video_file: str | PathLike,
    audio_file: str | PathLike,
    dst_file: str | PathLike,
    subtitle_file: str | PathLike | None = None,
) -> None:
    """Merges a video file, an audio file, and optionally a subtitle file into a new video file.

    It uses the ffmpeg library for media processing. The resulting video will have the audio
    and optionally the subtitles embedded.

    Args:
        video_file:
            A ``str`` or ``PathLike`` object representing the path to the source video file.
        audio_file:
            A ``str`` or ``PathLike`` object representing the path to the audio file to be added.
        dst_file:
            A ``str`` or ``PathLike`` object representing the path to the destination video file.
        subtitle_file:
            A ``str``, ``PathLike``, or ``None`` representing the path to
            the subtitle file to be added. Default is ``None``.
    """
    import ffmpeg
    kwargs = {"vf": f"ass={str(subtitle_file)}"} if subtitle_file is not None else {'vcodec': 'copy'}
    video_input = ffmpeg.input(video_file)
    audio_input = ffmpeg.input(audio_file)
    output = ffmpeg.output(
        video_input.video,
        audio_input.audio,
        dst_file,
        **kwargs,
        acodec="aac",
        ab="128k",
    )
    output.run(overwrite_output=True)


def _csscolor_to_rgb(css_name: str) -> tuple[int, int, int]:
    color_dict = {
        'aliceblue': '#F0F8FF',
        'antiquewhite': '#FAEBD7',
        'aqua': '#00FFFF',
        'aquamarine': '#7FFFD4',
        'azure': '#F0FFFF',
        'beige': '#F5F5DC',
        'bisque': '#FFE4C4x',
        'black': '#000000',
        'blanchedalmond': '#FFEBCD',
        'blue': '#0000FF',
        'blueviolet': '#8A2BE2',
        'brown': '#A52A2A',
        'burlywood': '#DEB887',
        'cadetblue': '#5F9EA0',
        'chartreuse': '#7FFF00',
        'chocolate': '#D2691E',
        'coral': '#FF7F50',
        'cornflowerblue': '#6495ED',
        'cornsilk': '#FFF8DC',
        'crimson': '#DC143C',
        'cyan': '#00FFFF',
        'darkblue': '#00008B',
        'darkcyan': '#008B8B',
        'darkgoldenrod': '#B8860B',
        'darkgray': '#A9A9A9',
        'darkgrey': '#A9A9A9',
        'darkgreen': '#006400',
        'darkkhaki': '#BDB76B',
        'darkmagenta': '#8B008B',
        'darkolivegreen': '#556B2F',
        'darkorange': '#FF8C00',
        'darkorchid': '#9932CC',
        'darkred': '#8B0000',
        'darksalmon': '#E9967A',
        'darkseagreen': '#8FBC8F',
        'darkslateblue': '#483D8B',
        'darkslategray': '#2F4F4F',
        'darkslategrey': '#2F4F4F',
        'darkturquoise': '#00CED1',
        'darkviolet': '#9400D3',
        'deeppink': '#FF1493',
        'deepskyblue': '#00BFFF',
        'dimgray': '#696969',
        'dimgrey': '#696969',
        'dodgerblue': '#1E90FF',
        'firebrick': '#B22222',
        'floralwhite': '#FFFAF0',
        'forestgreen': '#228B22',
        'fuchsia': '#FF00FF',
        'gainsboro': '#DCDCDC',
        'ghostwhite': '#F8F8FF',
        'gold': '#FFD700',
        'goldenrod': '#DAA520',
        'gray': '#808080',
        'grey': '#808080',
        'green': '#008000',
        'greenyellow': '#ADFF2F',
        'honeydew': '#F0FFF0',
        'hotpink': '#FF69B4',
        'indianred': '#CD5C5C',
        'indigo': '#4B0082',
        'ivory': '#FFFFF0',
        'khaki': '#F0E68C',
        'lavender': '#E6E6FA',
        'lavenderblush': '#FFF0F5',
        'lawngreen': '#7CFC00',
        'lemonchiffon': '#FFFACD',
        'lightblue': '#ADD8E6',
        'lightcoral': '#F08080',
        'lightcyan': '#E0FFFF',
        'lightgoldenrodyellow': '#FAFAD2',
        'lightgray': '#D3D3D3',
        'lightgrey': '#D3D3D3',
        'lightgreen': '#90EE90',
        'lightpink': '#FFB6C1',
        'lightsalmon': '#FFA07A',
        'lightseagreen': '#20B2AA',
        'lightskyblue': '#87CEFA',
        'lightslategray': '#778899',
        'lightslategrey': '#778899',
        'lightsteelblue': '#B0C4DE',
        'lightyellow': '#FFFFE0',
        'lime': '#00FF00',
        'limegreen': '#32CD32',
        'linen': '#FAF0E6',
        'magenta': '#FF00FF',
        'maroon': '#800000',
        'mediumaquamarine': '#66CDAA',
        'mediumblue': '#0000CD',
        'mediumorchid': '#BA55D3',
        'mediumpurple': '#9370DB',
        'mediumseagreen': '#3CB371',
        'mediumslateblue': '#7B68EE',
        'mediumspringgreen': '#00FA9A',
        'mediumturquoise': '#48D1CC',
        'mediumvioletred': '#C71585',
        'midnightblue': '#191970',
        'mintcream': '#F5FFFA',
        'mistyrose': '#FFE4E1',
        'moccasin': '#FFE4B5',
        'navajowhite': '#FFDEAD',
        'navy': '#000080',
        'oldlace': '#FDF5E6',
        'olive': '#808000',
        'olivedrab': '#6B8E23',
        'orange': '#FFA500',
        'orangered': '#FF4500',
        'orchid': '#DA70D6',
        'palegoldenrod': '#EEE8AA',
        'palegreen': '#98FB98',
        'paleturquoise': '#AFEEEE',
        'palevioletred': '#DB7093',
        'papayawhip': '#FFEFD5',
        'peachpuff': '#FFDAB9',
        'peru': '#CD853F',
        'pink': '#FFC0CB',
        'plum': '#DDA0DD',
        'powderblue': '#B0E0E6',
        'purple': '#800080',
        'rebeccapurple': '#663399',
        'red': '#FF0000',
        'rosybrown': '#BC8F8F',
        'royalblue': '#4169E1',
        'saddlebrown': '#8B4513',
        'salmon': '#FA8072',
        'sandybrown': '#F4A460',
        'seagreen': '#2E8B57',
        'seashell': '#FFF5EE',
        'sienna': '#A0522D',
        'silver': '#C0C0C0',
        'skyblue': '#87CEEB',
        'slateblue': '#6A5ACD',
        'slategray': '#708090',
        'slategrey': '#708090',
        'snow': '#FFFAFA',
        'springgreen': '#00FF7F',
        'steelblue': '#4682B4',
        'tan': '#D2B48C',
        'teal': '#008080',
        'thistle': '#D8BFD8',
        'tomato': '#FF6347',
        'turquoise': '#40E0D0',
        'violet': '#EE82EE',
        'wheat': '#F5DEB3',
        'white': '#FFFFFF',
        'whitesmoke': '#F5F5F5',
        'yellow': '#FFFF00',
        'yellowgreen': '#9ACD32',
    }
    if css_name in color_dict:
        hex_str = color_dict[css_name]
        return _hex_to_rgb(hex_str)
    else:
        raise ValueError(f'Invalid color name: {css_name}')


def _hex_to_rgb(hex_color: str) -> tuple[int, int, int]:
    hex_color = hex_color.lstrip('#')
    r = int(hex_color[0:2], 16)
    g = int(hex_color[2:4], 16)
    b = int(hex_color[4:6], 16)
    return (r, g, b)


def to_rgb(color: str | tuple[int, int, int] | Sequence[int]) -> tuple[int, int, int]:
    """Converts a color input to its RGB tuple representation.

    The input can be either a string representing a color name or a hex code,
    or a sequence of integers representing the RGB values.

    Args:
        color:
            A union type that can be either a string or a tuple containing RGB integers or a sequence of integers.
            If ``str``: The string can be either a CSS color name (e.g., 'red') or a hexadecimal RGB string
            (e.g., ``'#FF0000'``). If ``tuple[int, int, int]`` or ``Sequence[int]``, it represents RGB values
            as integers between 0 and 255 (`e.g.`, ``(255, 0, 0)``).

    Returns:
        A tuple of three integers ``(R, G, B)`` that represent the RGB values.
    """
    if isinstance(color, SequenceType) and all(isinstance(x, int) for x in color):
        return (int(color[0]), int(color[1]), int(color[2]))
    elif isinstance(color, str):
        if not color.startswith('#'):
            return _csscolor_to_rgb(color.lower())
        else:
            return _hex_to_rgb(color)
    else:
        raise ValueError



import numpy as np
import pytest

from movis.attribute import Attribute, AttributeType

attribute_params = [
    (0.5, AttributeType.SCALAR),
    (1.0, AttributeType.ANGLE),
    (0.3, AttributeType.COLOR),
    (0.2, AttributeType.VECTOR2D),
    (0.9, AttributeType.VECTOR3D),
]


@pytest.mark.parametrize("value, value_type", attribute_params)
def test_attribute_init(value: float, value_type: AttributeType):
    attr = Attribute(value, value_type)
    assert isinstance(attr.init_value, np.ndarray)
    if value_type == AttributeType.SCALAR or value_type == AttributeType.ANGLE:
        assert attr.init_value.shape == (1,)
        val = attr(0.0)
        assert val.shape == (1,)
    elif value_type == AttributeType.VECTOR2D:
        assert attr.init_value.shape == (2,)
        val = attr(0.0)
        assert val.shape == (2,)
    elif value_type == AttributeType.VECTOR3D or value_type == AttributeType.COLOR:
        assert attr.init_value.shape == (3,)
        val = attr(0.0)
        assert val.shape == (3,)
    assert isinstance(val, np.ndarray)
    assert val.dtype == np.float64


@pytest.mark.parametrize("value, value_type", attribute_params)
def test_attribute_init_range(value: float, value_type: AttributeType):
    attr = Attribute(value, value_type, range=(0.0, 0.1))
    assert isinstance(attr.init_value, np.ndarray)
    if value_type == AttributeType.SCALAR or value_type == AttributeType.ANGLE:
        assert attr.init_value.shape == (1,)
        val = attr(0.0)
        assert val.shape == (1,)
    elif value_type == AttributeType.VECTOR2D:
        assert attr.init_value.shape == (2,)
        val = attr(0.0)
        assert val.shape == (2,)
    elif value_type == AttributeType.VECTOR3D or value_type == AttributeType.COLOR:
        assert attr.init_value.shape == (3,)
        val = attr(0.0)
        assert val.shape == (3,)
    assert np.all(val == 0.1)



from __future__ import annotations

from typing import Hashable, Protocol

import numpy as np

AUDIO_SAMPLING_RATE = 44100
AUDIO_BLOCK_SIZE = 1024


class Layer(Protocol):
    """The protocol that defimes the minimal interface for a layer."""

    def __call__(self, time: float) -> np.ndarray | None:
        """The minimum required method to implement a layer. All layers must implement it.

        Specifically, this method returns a ``numpy.ndarray`` of shape ``(H, W, 4)`` with RGBA order and dtype as
        ``numpy.uint8``, or ``None``, given a time.
        When a ``numpy.ndarray`` is returned, Movis considers this array as an image and uses it as one of the layers
        for rendering the video. If ``None`` is returned, Movis does not render its layer.

        Args:
            time: A scalar variable representing time.

        Returns:
            ``None`` if nothing is to be rendered, otherwise ``numpy.ndarray``.
        """
        raise NotImplementedError


class BasicLayer(Protocol):
    """The protocol that defines the basic interface for a layer with some optional properties."""

    def __call__(self, time: float) -> np.ndarray | None:
        """The minimum required method to implement a layer. All layers must implement it.

        Specifically, this method returns a ``numpy.ndarray`` of shape ``(H, W, 4)`` with RGBA order and dtype as
        ``numpy.uint8``, or ``None``, given a time.
        When a ``numpy.ndarray`` is returned, Movis considers this array as an image and uses it as one of the layers
        for rendering the video. If ``None`` is returned, Movis does not render its layer.

        Args:
            time: A scalar variable representing time.

        Returns:
            ``None`` if nothing is to be rendered, otherwise ``numpy.ndarray``.
        """
        raise NotImplementedError

    @property
    def duration(self) -> float:
        """An optional but desirable property for any layer implementation.

        This property should return the duration for which the layer will persist.
        If not implemented, it is assumed that the layer has an indefinitely large duration.

        Returns:
            The duration for which the layer will persist.
        """
        return 1e6

    def get_key(self, time: float) -> Hashable:
        """An optional but desirable method for any layer implementation.

        This method returns a hashable value representing the 'state' of the layer at a given time.
        If the keys are the same, the array returned by this layer must also be identical.
        It is used for caching compositions, and in the case of videos where static frames persist,
        Movis will use the cache to accelerate video rendering.

        If not implemented, Movis assumes that the layer is independent at each time frame,
        `i.e.`, it will not use cache-based rendering.

        Returns:
            A hashable key that represents the state of the layer at the given time.
        """
        return time


class AudioLayer(Protocol):

    def __call__(self, time: float) -> np.ndarray | None:
        """The minimum required method to implement a layer. All layers must implement it.

        Specifically, this method returns a ``numpy.ndarray`` of shape ``(H, W, 4)`` with RGBA order and dtype as
        ``numpy.uint8``, or ``None``, given a time.
        When a ``numpy.ndarray`` is returned, Movis considers this array as an image and uses it as one of the layers
        for rendering the video. If ``None`` is returned, Movis does not render its layer.

        Args:
            time: A scalar variable representing time.

        Returns:
            ``None`` if nothing is to be rendered, otherwise ``numpy.ndarray``.
        """
        raise NotImplementedError

    def get_audio(self, start_time: float, end_time: float) -> np.ndarray | None:
        """An optional method for implementing an audio layer.

        This method returns an audio clip of the layer between the given start and end times.
        The returned audio clip should be a two-dimensional ``numpy.ndarray`` with a shape of ``(2,T)``,
        where ``T`` is the number of samples in the audio clip and ``2`` is the number of channels.
        The sample rate of the audio clip should be ``AUDIO_SAMPLING_RATE`` (= 44100).

        If not implemented, Movis assumes that the layer does not have any audio.

        Args:
            start_time: The start time of the audio clip.
            end_time: The end time of the audio clip.

        Returns:
            A two-dimensional ``numpy.ndarray`` with a shape of ``(2,T)`` and ``sample_rate``
            is an integer representing the sample rate of the audio clip.
        """
        raise NotImplementedError



from typing import Sequence

import numpy as np


class TimelineMixin:
    def __init__(
        self, start_times: Sequence[float], end_times: Sequence[float]
    ) -> None:
        assert len(start_times) == len(
            end_times
        ), f"{len(start_times)} != {len(end_times)}"
        self.start_times: np.ndarray = np.asarray(start_times, dtype=float)
        self.end_times: np.ndarray = np.asarray(end_times, dtype=float)

    def get_state(self, time: float) -> int:
        """Returns the index of the state at the given time, or ``-1`` if no state exists."""
        idx = np.searchsorted(self.end_times, time, side='right')
        if idx < len(self.start_times) and self.start_times[idx] <= time < self.end_times[idx]:
            return int(idx)
        else:
            return -1

    @property
    def duration(self):
        """Returns the duration of the timeline."""
        return self.end_times[-1] - self.start_times[0]



import numpy as np
import pytest

from movis.enum import BlendingMode, MatteMode
from movis.imgproc import alpha_composite

alpha_composite_params = [
    (0.5, BlendingMode.NORMAL),
    (0.2, BlendingMode.MULTIPLY),
    (1.0, BlendingMode.SCREEN),
    (0.9, BlendingMode.OVERLAY),
    (0.8, BlendingMode.DARKEN),
    (0.7, BlendingMode.LIGHTEN),
    (0.6, BlendingMode.COLOR_DODGE),
    (0.5, BlendingMode.LINEAR_DODGE),
    (0.4, BlendingMode.LINEAR_BURN),
    (0.3, BlendingMode.HARD_LIGHT),
    (0.1, BlendingMode.SOFT_LIGHT),
    (0.2, BlendingMode.VIVID_LIGHT),
    (0.3, BlendingMode.LINEAR_LIGHT),
    (0.4, BlendingMode.PIN_LIGHT),
    (0.5, BlendingMode.DIFFERENCE),
    (0.6, BlendingMode.EXCLUSION),
    (0.7, BlendingMode.SUBTRACT),
]


@pytest.mark.parametrize("opacity, blending_mode", alpha_composite_params)
def test_alpha_composite(opacity, blending_mode):
    bg = np.random.randint(0, 255, size=(128, 256, 4)).astype(np.uint8)
    fg = np.random.randint(0, 255, size=(64, 128, 4)).astype(np.uint8)

    for (x, y) in [(0, 0), (-10, -20), (96, 48), (-10, 48), (96, -20)]:
        bg_dst = alpha_composite(
            bg.copy(), fg, position=(x, y),
            opacity=opacity, blending_mode=blending_mode)
        assert bg.shape == bg_dst.shape
        assert bg_dst.dtype == np.uint8


@pytest.mark.parametrize("opacity, blending_mode", alpha_composite_params)
def test_alpha_composite_outside(opacity, blending_mode):
    bg = np.random.randint(0, 255, size=(128, 256, 4)).astype(np.uint8)
    fg = np.random.randint(0, 255, size=(64, 128, 4)).astype(np.uint8)
    for (x, y) in [(-200, -200), (512, 512)]:
        bg_dst = alpha_composite(
            bg.copy(), fg, position=(x, y),
            opacity=opacity, blending_mode=blending_mode)
        np.testing.assert_allclose(bg, bg_dst)
        assert bg_dst.dtype == np.uint8


@pytest.mark.parametrize("opacity, blending_mode", alpha_composite_params)
def test_alpha_composite_alphamatte(opacity, blending_mode):
    bg = np.random.randint(0, 255, size=(128, 256, 4)).astype(np.uint8)
    fg = np.random.randint(0, 255, size=(64, 128, 4)).astype(np.uint8)
    for (x, y) in [(0, 0), (-10, -20), (96, 48), (-10, 48), (96, -20)]:
        bg_dst = alpha_composite(
            bg.copy(), fg, position=(x, y),
            opacity=opacity, blending_mode=blending_mode, matte_mode=MatteMode.ALPHA)
        np.testing.assert_allclose(bg[:, :, 3], bg_dst[:, :, 3])
        assert bg_dst.dtype == np.uint8


@pytest.mark.parametrize("opacity, blending_mode", alpha_composite_params)
def test_alpha_composite_luminancematte(opacity, blending_mode):
    bg = np.random.randint(0, 255, size=(128, 256, 4)).astype(np.uint8)
    fg = np.random.randint(0, 255, size=(64, 128, 4)).astype(np.uint8)
    for (x, y) in [(0, 0), (-10, -20), (96, 48), (-10, 48), (96, -20)]:
        bg2 = bg.copy()
        bg2[:, :, :3] = 0
        bg_dst = alpha_composite(
            bg2, fg, position=(x, y),
            opacity=opacity, blending_mode=blending_mode, matte_mode=MatteMode.LUMINANCE)
        np.testing.assert_allclose(bg_dst[:, :, 3], 0)
        assert bg_dst.dtype == np.uint8



from __future__ import annotations

from typing import NamedTuple, Sequence

import numpy as np

from movis.enum import BlendingMode, Direction

from .attribute import Attribute, AttributeType


class TransformValue(NamedTuple):
    """A named tuple that encapsulates various transformation properties that can be applied to a layer.

    The properties of `TransformValue` include the anchor point, position, scale, rotation, opacity, and origin point.

    Attributes:
        anchor_point:
            A tuple of two floats representing the anchor point ``(x, y)`` of an object. Defaults to ``(0.0, 0.0)``.
        position:
            A tuple of two floats representing the position ``(x, y)`` of an object. Defaults to ``(0.0, 0.0)``.
        scale:
            A tuple of two floats representing the scale ``(x, y)`` of an object. Defaults to ``(1.0, 1.0)``.
        rotation:
            A float value representing the rotation angle in degrees. Defaults to ``0.0``.
        opacity:
            A float value representing the opacity of an object. Must be in the range ``[0, 1]``. Defaults to ``1.0``.
        origin_point:
            An enum value from Direction representing the origin point for transformations.
            Defaults to ``Direction.CENTER``.
        origin_point:
            An enum value from Direction representing the blending mode of the layer.
            Defaults to ``BlendingMode.NORMAL``.
    """

    anchor_point: tuple[float, float] = (0.0, 0.0)
    position: tuple[float, float] = (0.0, 0.0)
    scale: tuple[float, float] = (1.0, 1.0)
    rotation: float = 0.0
    opacity: float = 1.0
    origin_point: Direction = Direction.CENTER
    blending_mode: BlendingMode = BlendingMode.NORMAL


class Transform:
    """A class responsible for encapsulating the various transformation attributes for a layer.

    It utilizes ``Attribute`` class to enforce types and optionally ranges for each attribute.

    Args:
        position:
            A float, tuple of floats, or numpy ndarray representing the position ``(x, y)`` of an object.
            Defaults to ``(0.0, 0.0)``.
        scale:
            A float, tuple of floats, or numpy ndarray representing the scale ``(x, y)`` of an object.
            Defaults to ``(1.0, 1.0)``.
        rotation:
            A float value representing the rotation angle in degrees. Defaults to ``0.0``.
        opacity:
            A float value representing the opacity of an object. Must be in the range ``[0, 1]``. Defaults to ``1.0``.
        anchor_point:
            A float, tuple of floats, or ``numpy.ndarray`` representing the anchor point ``(x, y)`` of an object.
            Defaults to ``(0.0, 0.0)``.
        origin_point:
            An enum value from Direction or a string representing the origin point for transformations.
            Defaults to ``Direction.CENTER``.
    """

    def __init__(
        self,
        position: float | tuple[float, float] | np.ndarray = (0.0, 0.0),
        scale: float | tuple[float, float] | np.ndarray = (1.0, 1.0),
        rotation: float = 0.0,
        opacity: float = 1.0,
        anchor_point: float | tuple[float, float] | np.ndarray = (0.0, 0.0),
        origin_point: Direction | str = Direction.CENTER,
        blending_mode: BlendingMode | str = BlendingMode.NORMAL,
    ):
        self.position = Attribute(position, AttributeType.VECTOR2D)
        self.scale = Attribute(scale, AttributeType.VECTOR2D)
        self.rotation = Attribute(rotation, AttributeType.SCALAR)
        self.opacity = Attribute(opacity, AttributeType.SCALAR, range=(0., 1.))
        self.anchor_point = Attribute(anchor_point, AttributeType.VECTOR2D)
        self.origin_point = Direction.from_string(origin_point) if isinstance(origin_point, str) else origin_point
        self.blending_mode = BlendingMode.from_string(blending_mode) \
            if isinstance(blending_mode, str) else blending_mode

    @property
    def attributes(self) -> dict[str, Attribute]:
        return {
            "anchor_point": self.anchor_point,
            "position": self.position,
            "scale": self.scale,
            "rotation": self.rotation,
            "opacity": self.opacity,
        }

    @classmethod
    def from_positions(
        cls,
        size: tuple[int, int],
        top: float | None = None,
        bottom: float | None = None,
        left: float | None = None,
        right: float | None = None,
        object_fit: str | None = None
    ) -> "Transform":
        """Allows to create `Transform` by specifying the position based on the edges ``(top, bottom, left, right)``.

        The ``object_fit`` parameter specifies how the object should scale to fit the canvas or container.
        If ``object_fit='contain'``, the object will scale to fit within the canvas while preserving its aspect ratio.
        If ``'cover'``, the object will scale to completely cover the canvas, also preserving its aspect ratio.

        The method calculates the position and origin point of the object based on the supplied arguments
        and returns a new Transform object.

        Args:
            size:
                A tuple of two integers representing the width (``W``) and height (``H``) of the canvas or container.
            top:
                Optional float, distance from the top edge of the layer. Default is ``None``.
            bottom:
                Optional float, distance from the bottom edge of the layer. Default is ``None``.
            left:
                Optional float, distance from the left edge of the layer. Default is ``None``.
            right:
                Optional float, distance from the right edge of the layer. Default is ``None``.
            object_fit:
                Optional string, specifies the scaling behavior. Accepts either ``'contain'`` or ``'cover'``.
                Default is ``None`` (do nothing).

        Returns:
            A new `Transform` object with the specified position, scale, and origin point.
        """
        W, H = size[0], size[1]
        if top is None and bottom is None and left is None and right is None:
            x, y = W / 2, H / 2
            origin_point = Direction.CENTER
            if object_fit is None:
                scale = 1.0
            elif object_fit == "contain":
                scale = min(W, H) / max(W, H)
            elif object_fit == "cover":
                scale = max(W, H) / min(W, H)
            else:
                raise ValueError(f"Invalid object_fit: {object_fit}. contain or cover is expected.")
        else:
            if top is not None and bottom is None and left is None and right is None:
                x, y = W / 2, top
                origin_point = Direction.TOP_CENTER
            elif top is None and bottom is not None and left is None and right is None:
                x, y = W / 2, H - bottom
                origin_point = Direction.BOTTOM_CENTER
            elif top is None and bottom is None and left is not None and right is None:
                x, y = left, H / 2
                origin_point = Direction.CENTER_LEFT
            elif top is None and bottom is None and left is None and right is not None:
                x, y = W - right, H / 2
                origin_point = Direction.CENTER_RIGHT
            elif top is not None and bottom is None and left is not None and right is None:
                x, y = left, top
                origin_point = Direction.TOP_LEFT
            elif top is not None and bottom is None and left is None and right is not None:
                x, y = W - right, top
                origin_point = Direction.TOP_RIGHT
            elif top is None and bottom is not None and left is not None and right is None:
                x, y = left, H - bottom
                origin_point = Direction.BOTTOM_LEFT
            elif top is None and bottom is not None and left is None and right is not None:
                x, y = W - right, H - bottom
                origin_point = Direction.BOTTOM_RIGHT
            else:
                raise ValueError("Invalid pair of arguments")
            scale = 1.0
        return cls(position=(x, y), origin_point=origin_point, scale=scale)

    def get_current_value(self, layer_time: float) -> TransformValue:
        """Retrieves the current transformation attributes for a given time, encapsulated within a ``TransformValue``.

        This includes the anchor point, position, scale, rotation, and opacity, all evaluated at ``layer_time``.
        The transformation attributes are converted to their corresponding 2D vectors or scalar values as appropriate,
        for easy use or manipulation.

        Args:
            layer_time:
                A float representing the time for which you want to get the current value of
                all the transformation attributes.

        Returns:
            ``TransformValue`` that encapsulates the current transformation properties
            (anchor point, position, scale, rotation, opacity, and origin point) for the given ``layer_time``.
        """
        return TransformValue(
            anchor_point=transform_to_2dvector(self.anchor_point(layer_time)),
            position=transform_to_2dvector(self.position(layer_time)),
            scale=transform_to_2dvector(self.scale(layer_time)),
            rotation=transform_to_1dscalar(self.rotation(layer_time)),
            opacity=transform_to_1dscalar(self.opacity(layer_time)),
            origin_point=self.origin_point,
            blending_mode=self.blending_mode,
        )

    def __repr__(self) -> str:
        return f"Transform(ap={self.anchor_point}, pos={self.position}, " \
            f"s={self.scale}, rot={self.rotation}, op={self.opacity}, blend={self.blending_mode})"


def transform_to_1dscalar(x: float | Sequence[float] | np.ndarray) -> float:
    """Transform a scalar, tuple, list or numpy array to a scalar.

    Args:
        x:
            The value to transform.

    Returns:
        A scalar.
    """
    if isinstance(x, float):
        return x
    elif isinstance(x, np.ndarray) and x.shape == ():
        return float(x)
    else:
        if 0 < len(x):
            return float(x[0])
        else:
            raise ValueError(f"Invalid value: {x}")


def transform_to_2dvector(
    x: float | Sequence[float] | np.ndarray
) -> tuple[float, float]:
    """Transform a scalar, tuple, list or numpy array to a 2D vector.

    Args:
        x:
            The value to transform.

    Returns:
        A 2D vector.
    """
    if isinstance(x, float):
        return (x, x)
    elif isinstance(x, np.ndarray) and x.shape == ():
        return (float(x), float(x))
    elif len(x) == 1:
        return (float(x[0]), float(x[0]))
    elif len(x) == 2:
        return (float(x[0]), float(x[1]))
    else:
        raise ValueError(f"Invalid value: {x}")


def transform_to_3dvector(
    x: float | Sequence[float] | np.ndarray
) -> tuple[float, float, float]:
    """Transform a scalar, tuple, list or numpy array to a 3D vector.

    Args:
        x:
            The value to transform.

    Returns:
        A 3D vector.
    """
    if isinstance(x, float):
        return (x, x, x)
    elif isinstance(x, np.ndarray) and x.shape == ():
        return (float(x), float(x), float(x))
    elif len(x) == 1:
        y = float(x[0])
        return (y, y, y)
    elif len(x) == 3:
        return (float(x[0]), float(x[1]), float(x[2]))
    else:
        raise ValueError(f"Invalid value: {x}")



from __future__ import annotations

import bisect
from typing import Any, Callable, Sequence

import numpy as np

from .enum import AttributeType, Easing


def linear(t: float) -> float:
    return t


def flat(t: float) -> float:
    return 0.0


class EaseIn:
    def __init__(self, n: int):
        self.n = n

    def __call__(self, t: float) -> float:
        return t ** self.n


class EaseOut:
    def __init__(self, n: int):
        self.n = n

    def __call__(self, t: float) -> float:
        return 1.0 - (1.0 - t) ** self.n


class EaseInOut:
    def __init__(self, n: int):
        self.n = n

    def __call__(self, t: float) -> float:
        return 0.5 * (2 * t) ** self.n if t < 0.5 else 1.0 - 0.5 * (1.0 - 2 * (t - 0.5)) ** self.n


EASING_TO_FUNC: dict[Easing, Callable[[float], float]] = {  # type: ignore
    Easing.LINEAR: linear,
    Easing.EASE_IN: EaseIn(2),
    Easing.EASE_OUT: EaseOut(2),
    Easing.EASE_IN_OUT: EaseInOut(2),
    Easing.FLAT: flat,
    Easing.EASE_IN2: EaseIn(2),
    Easing.EASE_IN3: EaseIn(3),
    Easing.EASE_IN4: EaseIn(4),
    Easing.EASE_IN5: EaseIn(5),
    Easing.EASE_IN6: EaseIn(6),
    Easing.EASE_IN7: EaseIn(7),
    Easing.EASE_IN8: EaseIn(8),
    Easing.EASE_IN9: EaseIn(9),
    Easing.EASE_IN10: EaseIn(10),
    Easing.EASE_IN12: EaseIn(12),
    Easing.EASE_IN14: EaseIn(14),
    Easing.EASE_IN16: EaseIn(16),
    Easing.EASE_IN18: EaseIn(18),
    Easing.EASE_IN20: EaseIn(20),
    Easing.EASE_IN25: EaseIn(25),
    Easing.EASE_IN30: EaseIn(30),
    Easing.EASE_IN35: EaseIn(35),
    Easing.EASE_OUT2: EaseOut(2),
    Easing.EASE_OUT3: EaseOut(3),
    Easing.EASE_OUT4: EaseOut(4),
    Easing.EASE_OUT5: EaseOut(5),
    Easing.EASE_OUT6: EaseOut(6),
    Easing.EASE_OUT7: EaseOut(7),
    Easing.EASE_OUT8: EaseOut(8),
    Easing.EASE_OUT9: EaseOut(9),
    Easing.EASE_OUT10: EaseOut(10),
    Easing.EASE_OUT12: EaseOut(12),
    Easing.EASE_OUT14: EaseOut(14),
    Easing.EASE_OUT16: EaseOut(16),
    Easing.EASE_OUT18: EaseOut(18),
    Easing.EASE_OUT20: EaseOut(20),
    Easing.EASE_OUT25: EaseOut(25),
    Easing.EASE_OUT30: EaseOut(30),
    Easing.EASE_OUT35: EaseOut(35),
    Easing.EASE_IN_OUT2: EaseInOut(2),
    Easing.EASE_IN_OUT3: EaseInOut(3),
    Easing.EASE_IN_OUT4: EaseInOut(4),
    Easing.EASE_IN_OUT5: EaseInOut(5),
    Easing.EASE_IN_OUT6: EaseInOut(6),
    Easing.EASE_IN_OUT7: EaseInOut(7),
    Easing.EASE_IN_OUT8: EaseInOut(8),
    Easing.EASE_IN_OUT9: EaseInOut(9),
    Easing.EASE_IN_OUT10: EaseInOut(10),
    Easing.EASE_IN_OUT12: EaseInOut(12),
    Easing.EASE_IN_OUT14: EaseInOut(14),
    Easing.EASE_IN_OUT16: EaseInOut(16),
    Easing.EASE_IN_OUT18: EaseInOut(18),
    Easing.EASE_IN_OUT20: EaseInOut(20),
    Easing.EASE_IN_OUT25: EaseInOut(25),
    Easing.EASE_IN_OUT30: EaseInOut(30),
    Easing.EASE_IN_OUT35: EaseInOut(35),
}


class Motion:
    """Defines a motion of ``Attribute`` instances used for keyframe animations.

    This instance is basically initialized by calling ``enable_motion()`` in ``Attribute`` instances.
    keyframes, values and correspoing motion types can be added using ``append()`` or ``extend()``.
    By using these keyframes, ``Motion`` instance returns the complemented values at the given time.

    Args:
        init_value:
            The initial value to use. It is used when no animation is set,
            or when an animation with zero keyframes is set.
        value_type:
            Specifies the type of value that the ``Attribute`` will handle.

    Examples:
        >>> import movis as mv
        >>> attr = mv.Attribute(1.0, mv.AttributeType.SCALAR)
        >>> attr.enable_motion().append(0.0, 0.0).append(1.0, 1.0)
        >>> len(attr.motion)
        2
        >>> attr.enable_motion().extend([2.0, 3.0], [2.0, 3.0], ['ease_in_out'])
        >>> len(attr.motion)
        4
        >>> attr.motion.clear()
        >>> len(attr.motion)
        0
        >>> assert attr(0.0) == attr.init_value
    """

    def __init__(
        self,
        init_value: float | Sequence[float] | np.ndarray | None = None,
        value_type: AttributeType = AttributeType.SCALAR,
    ):
        self.keyframes: list[float] = []
        self.values: list[np.ndarray] = []
        self.easings: list[Callable[[float], float]] = []
        self.init_value: np.ndarray | None = transform_to_numpy(init_value, value_type) \
            if init_value is not None else None
        self.value_type = value_type

    def __len__(self) -> int:
        return len(self.keyframes)

    def __call__(self, prev_value: np.ndarray, layer_time: float) -> np.ndarray:
        if len(self.keyframes) == 0:
            if self.init_value is not None:
                return self.init_value
            raise ValueError("No keyframes")
        elif len(self.keyframes) == 1:
            return self.values[0]

        if layer_time < self.keyframes[0]:
            return self.values[0]
        elif self.keyframes[-1] <= layer_time:
            return self.values[-1]
        else:
            i = bisect.bisect(self.keyframes, layer_time)
            m, M = self.values[i - 1], self.values[i]
            duration = self.keyframes[i] - self.keyframes[i - 1]
            t = (layer_time - self.keyframes[i - 1]) / duration
            t = self.easings[i - 1](t)
            return m + (M - m) * t

    def append(
        self,
        keyframe: float,
        value: float | Sequence[float] | np.ndarray,
        easing: str | Easing | Callable[[float], float] = Easing.LINEAR,
    ) -> "Motion":
        """Append a single keyframe.

        Args:
            keyframe:
                time of the keyframe.
            value:
                value of the keyframe.
            easing:
                motion type of the keyframe. This must be a string,
                ``Easing`` enum, or an easing function
                ``f: float -> float`` that satisfies ``f(0) == 0`` and ``f(1) == 1``.
                The default is ``Easing.LINEAR`` (linear completion).
        """
        i = bisect.bisect(self.keyframes, keyframe)
        if 0 < i and self.keyframes[i - 1] == keyframe:
            raise ValueError(f"Keyframe {keyframe} already exists")
        self.keyframes.insert(i, float(keyframe))
        self.values.insert(i, transform_to_numpy(value, self.value_type))
        if isinstance(easing, str):
            easing_func = EASING_TO_FUNC[Easing.from_string(easing)]
        elif isinstance(easing, Easing):
            easing_func = EASING_TO_FUNC[easing]
        elif callable(easing):
            easing_func = easing
        else:
            raise ValueError(f"Invalid easing type: {type(easing)}")
        self.easings.insert(i, easing_func)
        return self

    def extend(
        self,
        keyframes: Sequence[float],
        values: Sequence[float | Sequence[float] | np.ndarray],
        easings: Sequence[str | Easing | Callable[[float], float]] | None = None,
    ) -> "Motion":
        """Append multiple keyframes.

        Args:
            keyframes:
                times of the keyframes.
            values:
                values of the keyframes. The length of ``values`` must be the same as ``keyframes``.
            easings:
                motion types of the keyframes. Each element of ``easings`` must be a string, ``Easing`` enum,
                or an easing function ``f: float -> float`` that satisfies ``f(0) == 0`` and ``f(1) == 1``.
                Note that the length of ``easings`` must be the same as ``len(keyframes)`` or ``len(keyframes) - 1``.
                If ``easings`` is ``None``, ``Easing.LINEAR`` is used for all keyframes.
                If ``len(easings) == len(keyframes) - 1``, ``Motion`` automatically adds
                ``Easing.LINEAR`` to the end of ``easings``.

        Examples:
            >>> import movis as mv
            >>> motion = mv.Motion(value_type=mv.AttributeType.SCALAR)
            >>> # add two keyframes and The type of motion for that period is ``Easing.EASE_IN_OUT``
            >>> motion.extend(keyframes=[0.0, 1.0], values=[0.0, 1.0], easings=['ease_in_out'])
            >>> # add other two keyframes and The type of motion for that period is ``Easing.LINEAR``
            >>> motion.extend([2.0, 3.0], [2.0, 3.0])
            >>> len(motion)
            4
        """
        assert len(keyframes) == len(values)
        if isinstance(easings, str) or isinstance(easings, Easing):
            raise ValueError("easings must be a list of strings or Easing enums")
        if easings is not None:
            if len(easings) == len(keyframes) - 1:
                easings = list(easings) + [Easing.LINEAR]
            assert len(keyframes) == len(easings)
        easings = (
            ["linear"] * len(keyframes) if easings is None else easings
        )
        converted_keyframes = [float(k) for k in keyframes]

        # Check if given keyframes already exist
        seen = set(self.keyframes)
        for keyframe in converted_keyframes:
            if keyframe in seen:
                raise ValueError(f"Keyframe {keyframe} already exists")
            seen.add(keyframe)

        updated_keyframes: list[float] = self.keyframes + converted_keyframes
        converted_values = [
            transform_to_numpy(v, self.value_type) for v in values]
        updated_values: list[np.ndarray[Any, np.dtype[np.float64]]] = self.values + converted_values

        def convert(t: str | Easing | Callable[[float], float]) -> Callable[[float], float]:
            if callable(t):
                return t
            elif isinstance(t, Easing):
                return EASING_TO_FUNC[t]
            elif isinstance(t, str):
                return EASING_TO_FUNC[Easing.from_string(t)]
            else:
                raise ValueError(f"Invalid easing type: {type(t)}")

        converted_easings = [convert(t) for t in easings]
        updated_easings: list[Callable[[float], float]] = self.easings + converted_easings

        zipped = sorted(zip(updated_keyframes, updated_values, updated_easings))
        keyframes_sorted, values_sorted, easings_sorted = zip(*zipped)
        self.keyframes = list(keyframes_sorted)
        self.values = list(values_sorted)
        self.easings = list(easings_sorted)
        return self

    def clear(self) -> "Motion":
        self.keyframes = []
        self.values = []
        self.easings = []
        return self


def transform_to_numpy(value: int | float | Sequence[float] | np.ndarray, value_type: AttributeType) -> np.ndarray:
    """Transform a scalar, tuple, list or numpy array to a numpy array.

    Args:
        value:
            The value to transform.
        value_type:
            The type of value that the attribute handles.

    Returns:
        A numpy array with the length of 1, 2 or 3.
        This length is determined by the ``value_type``.

    Examples:
        >>> from movis.motion import transform_to_numpy
        >>> transform_to_numpy(1.0, mv.AttributeType.SCALAR)
        array([1.])
        >>> transform_to_numpy([1.0, 2.0], mv.AttributeType.VECTOR2D)
        array([1., 2.])
        >>> transform_to_numpy(1.0, mv.AttributeType.VECTOR3D)
        array([1., 1., 1.])
    """
    if isinstance(value, (int, float)):
        if value_type in (AttributeType.SCALAR, AttributeType.ANGLE):
            return np.array([value], dtype=np.float64)
        elif value_type == AttributeType.VECTOR2D:
            return np.array([value, value], dtype=np.float64)
        elif value_type in (AttributeType.VECTOR3D, AttributeType.COLOR):
            return np.array([value, value, value], dtype=np.float64)
    elif isinstance(value, (Sequence, np.ndarray)):
        if len(value) == 2 and value_type == AttributeType.VECTOR2D or \
                len(value) == 3 and value_type in (AttributeType.VECTOR3D, AttributeType.COLOR) or \
                len(value) == 1 and value_type in (AttributeType.SCALAR, AttributeType.ANGLE):
            return np.array(value, dtype=np.float64)
        else:
            raise ValueError(f"Invalid value type: {value_type}")
    raise ValueError(f"Invalid value type: {value_type}")



from __future__ import annotations

import cv2
import numpy as np

from ..attribute import Attribute, AttributesMixin, AttributeType
from ..util import to_rgb


class FillColor(AttributesMixin):
    """Fill the image with the specified color while preserving the alpha channel.

    Args:
        color:
            Color to fill. It can be specified as a tuple of RGB values or a string of color name.

    Animatable Attributes:
        ``color``
    """

    def __init__(self, color: tuple[int, int, int] | str = (255, 255, 255)):
        c = to_rgb(color)
        self.color = Attribute(c, AttributeType.COLOR, range=(0., 255.))

    def __call__(self, prev_image: np.ndarray, time: float) -> np.ndarray:
        assert prev_image.ndim == 3
        assert prev_image.shape[2] == 4, f'prev_image must be RGBA image, but {prev_image.shape}'
        rgb_image = np.full_like(
            prev_image[:, :, :3], np.round(self.color(time)).astype(np.uint8))
        alpha_image = prev_image[:, :, 3:]
        return np.concatenate([rgb_image, alpha_image], axis=2)


class HSLShift(AttributesMixin):
    """Shift hue, saturation, and luminance of the image.

    Args:
        hue:
            Hue shift in degrees.
        saturation:
            Saturation shift in the range ``[-1, 1]``.
        luminance:
            Luminance shift in the range ``[-1, 1]``.

    Animatable Attributes:
        ``hue``
        ``saturation``
        ``luminance``
    """

    def __init__(self, hue: float = 0.0, saturation: float = 0.0, luminance: float = 0.0):
        self.hue = Attribute(hue, AttributeType.SCALAR)
        self.saturation = Attribute(saturation, AttributeType.SCALAR, range=(-1., 1.))
        self.luminance = Attribute(luminance, AttributeType.SCALAR, range=(-1., 1.))

    def __call__(self, prev_image: np.ndarray, time: float) -> np.ndarray:
        assert prev_image.ndim == 3
        assert prev_image.shape[2] == 4, f'prev_image must be RGBA image, but {prev_image.shape}'
        hsv_image = cv2.cvtColor(prev_image[:, :, :3], cv2.COLOR_RGB2HSV).astype(np.float32)
        h, s, v = hsv_image[:, :, 0], hsv_image[:, :, 1], hsv_image[:, :, 2]
        dh = np.float32(self.hue(time))
        ds = np.float32(self.saturation(time))
        dl = np.float32(self.luminance(time))
        h = np.round((h + dh / 2) % 180).astype(np.uint8)
        s = np.clip(np.round(s + ds * 255), 0, 255).astype(np.uint8)
        v = np.clip(np.round(v + dl * 255), 0, 255).astype(np.uint8)
        new_hsv_image = cv2.cvtColor(np.stack([h, s, v], axis=2), cv2.COLOR_HSV2RGB)
        alpha_image = prev_image[:, :, 3:]
        return np.concatenate([new_hsv_image, alpha_image], axis=2)



from typing import Hashable, Protocol

import numpy as np


class Effect(Protocol):
    """The protocol that defimes the minimal interface for an effect."""

    def __call__(self, prev_image: np.ndarray, time: float) -> np.ndarray:
        """The minimum required method to implement an effect. All effects must implement it.

        Specifically, given a previous image described as ``prev_image`` and a time,
        it returns a ``numpy.ndarray`` of shape ``(H, W, 4)`` with RGBA order and dtype as
        ``numpy.uint8`` given a time.

        Args:
            prev_time:
                A ``numpy.ndarray`` of shape ``(H, W, 4)`` with RGBA order and dtype as ``numpy.uint8``.
            time:
                A scalar variable representing time.

        Returns:
            ``numpy.ndarray`` of shape ``(H, W, 4)`` with RGBA order and dtype as ``numpy.uint8``.
        """
        raise NotImplementedError


class BasicEffect(Effect):
    """The protocol that defines the basic interface for an effect with some methods."""

    def get_key(self, time: float) -> Hashable:
        """An optional but desirable method for any layer implementation.

        This method returns a hashable value representing the 'state' of the effect at a given time.
        If the keys are the same and given images are identical,
        the array returned by this layer must also be identical.
        It is used for caching compositions, and in the case of videos where static frames persist,
        Movis will use the cache to accelerate video rendering.

        If not implemented, Movis assumes that the layer is independent at each time frame,
        `i.e.`, it will not use cache-based rendering.

        Returns:
            A hashable key that represents the state of the layer at the given time.
        """
        return time



from .composition import Composition, LayerItem  # noqa
from .drawing import (Ellipse, FillProperty, Line, Rectangle,  # noqa
                      StrokeProperty, Text)
from .layer_ops import AlphaMatte, LuminanceMatte  # noqa
from .media import Audio, AudioSequence, Image, ImageSequence, Video  # noqa
from .mixin import TimelineMixin  # noqa
from .protocol import BasicLayer, Layer  # noqa
from .texture import Gradient, Stripe  # noqa



from __future__ import annotations

from typing import Hashable

import numpy as np

from ..attribute import Attribute, AttributesMixin, AttributeType
from ..enum import BlendingMode, MatteMode
from ..imgproc import alpha_composite
from .protocol import BasicLayer


class AlphaMatte(AttributesMixin):
    """A layer that applies alpha matte to the target layer using the mask layer.

    Alpha Matte is a algorihtm that overlays the target layer on the mask layer without changing the alpha channel.
    The mask layer and the target layer should have the same size and the same duration.
    Using the `Composition` layer is preferred if users want to align them.

    Args:
        mask:
            the base mask layer used for alpha matte. ``mask`` must comply with the ``BasicLayer`` protocol.
        target:
            the target layer to which alpha matte is applied. ``target`` must comply with the ``BasicLayer`` protocol.
        opacity:
            the opacity of the target layer. Defaults to ``1.0``.
        blending_mode:
            the blending mode of the target layer. Defaults to ``BlendingMode.NORMAL``.

    Animatable Attributes:
        ``opacity``
    """

    def __init__(
            self, mask: BasicLayer, target: BasicLayer,
            opacity: float = 1.0, blending_mode: BlendingMode | str = BlendingMode.NORMAL):
        self.mask = mask
        self.target = target
        self.opacity = Attribute(opacity, value_type=AttributeType.SCALAR, range=(0., 1.))
        self.blending_mode = BlendingMode.from_string(blending_mode) \
            if isinstance(blending_mode, str) else blending_mode

    def get_key(self, time: float) -> tuple[Hashable, Hashable, Hashable]:
        """Get the state for the given time."""
        attr_key = super().get_key(time)
        mask_key = self.mask.get_key(time) if hasattr(self.mask, 'get_key') else time
        target_key = self.target.get_key(time) if hasattr(self.target, 'get_key') else time
        return (attr_key, mask_key, target_key)

    @property
    def duration(self) -> float:
        """The duration of the layer."""
        return self.mask.duration

    def __call__(self, time: float) -> np.ndarray | None:
        if time < 0 or self.duration <= time:
            return None
        mask_frame = self.mask(time)
        if mask_frame is None:
            return None
        target_frame = self.target(time)
        if target_frame is None:
            return mask_frame
        opacity = float(self.opacity(time))
        return alpha_composite(
            mask_frame, target_frame, opacity=opacity,
            blending_mode=self.blending_mode, matte_mode=MatteMode.ALPHA)


class LuminanceMatte:
    """A layer that replaces the alpha channel of the target layer with the luminance of the mask layer.

    .. note::
        The mask layer and the target layer should have the same size and the same duration.
        Using the `Composition` layer is preferred if users want to align them.

    Args:
        mask:
            the base mask layer used for luminance matte. ``mask`` must comply with the ``BasicLayer`` protocol.
        target:
            the target layer to which luminance matte is applied.
            ``target`` must comply with the ``BasicLayer`` protocol.
    """

    def __init__(self, mask: BasicLayer, target: BasicLayer):
        self.mask = mask
        self.target = target

    def get_key(self, time: float) -> tuple[Hashable, Hashable]:
        """Get the state for the given time."""
        mask_key = self.mask.get_key(time) if hasattr(self.mask, 'get_key') else time
        target_key = self.target.get_key(time) if hasattr(self.target, 'get_key') else time
        return (mask_key, target_key)

    @property
    def duration(self) -> float:
        """The duration of the layer."""
        return self.mask.duration

    def __call__(self, time: float) -> np.ndarray | None:
        if time < 0 or self.duration <= time:
            return None
        mask_frame = self.mask(time)
        if mask_frame is None:
            return None
        target_frame = self.target(time)
        if target_frame is None:
            return None
        return alpha_composite(mask_frame, target_frame, matte_mode=MatteMode.LUMINANCE)



from __future__ import annotations

import numpy as np
from PySide6.QtGui import (QColor, QImage, QLinearGradient, QPainter,
                           QRadialGradient)

from ..attribute import Attribute, AttributesMixin, AttributeType
from ..imgproc import qimage_to_numpy
from ..util import to_rgb


class Gradient(AttributesMixin):
    """A layer that generates a gradient image.

    Args:
        size:
            the size of the generated image. Defaults to ``(100, 100)``.
        start_point:
            the start point of the gradient. Defaults to ``(0., 0.)``.
        end_point:
            the end point of the gradient. Defaults to ``(100., 100.)``.
        start_color:
            the start color of the gradient. Defaults to ``(0, 0, 0)``.
            the color can be specified as a tuple of ``(R, G, B)``
            or a string (`e.g.,` ``"#ff0000"``, or ``"red"``).
        end_color:
            the end color of the gradient. Defaults to ``(255, 255, 255)``.
        gradient_type:
            the type of the gradient. "linear" or "radial" is expected. Defaults to ``"linear"``.
        duration:
            the duration of the layer. Defaults to ``1e6``.

    Animatable Attributes:
        ``start_point``
        ``end_point``
        ``start_color``
        ``end_color``
    """

    def __init__(
        self,
        size: tuple[int, int] = (100, 100),
        start_point: tuple[float, float] = (0., 0.),
        end_point: tuple[float, float] = (100., 100.,),
        start_color: tuple[int, int, int] | str = (0, 0, 0),
        end_color: tuple[int, int, int] | str = (255, 255, 255),
        gradient_type: str = 'linear',
        duration: float = 1e6,
    ) -> None:
        self.size = size
        self.duration = duration
        self.start_point = Attribute(start_point, AttributeType.VECTOR2D)
        self.end_point = Attribute(end_point, AttributeType.VECTOR2D)
        cs = to_rgb(start_color)
        ce = to_rgb(end_color)
        self.start_color = Attribute(cs, AttributeType.COLOR, range=(0., 255.))
        self.end_color = Attribute(ce, AttributeType.COLOR, range=(0., 255.))
        if gradient_type not in ('linear', 'radial'):
            raise ValueError(f"Invalid gradation_type: {gradient_type}. 'linear' or 'radial' is expected.")
        self.gradient_type = gradient_type

    def __call__(self, time: float) -> np.ndarray | None:
        if time < 0 or time >= self.duration:
            return None
        width, height = self.size
        image = QImage(width, height, QImage.Format.Format_ARGB32)
        painter = QPainter(image)
        ps = self.start_point(time)
        pe = self.end_point(time)
        cs = np.round(self.start_color(time)).astype(int)
        ce = np.round(self.end_color(time)).astype(int)
        if self.gradient_type == 'linear':
            grad_linear = QLinearGradient(
                float(ps[0]) + 0.5, float(ps[1]) + 0.5, float(pe[0]) - 0.5, float(pe[1]) - 0.5)
            grad_linear.setColorAt(0, QColor(cs[2], cs[1], cs[0], 255))
            grad_linear.setColorAt(1, QColor(ce[2], ce[1], ce[0], 255))
            painter.fillRect(0, 0, width, height, grad_linear)
        elif self.gradient_type == 'radial':
            radius = np.sqrt(((ps - pe) ** 2).sum())
            grad_radial = QRadialGradient(float(ps[0]) + 0.5, float(ps[1]) + 0.5, float(radius))
            grad_radial.setColorAt(0, QColor(cs[2], cs[1], cs[0], 255))
            grad_radial.setColorAt(1, QColor(ce[2], ce[1], ce[0], 255))
            painter.fillRect(0, 0, width, height, grad_radial)
        painter.end()
        return qimage_to_numpy(image)


class Stripe(AttributesMixin):
    """A layer that generates a stripe pattern.

    Args:
        size:
            the size of the generated image. Defaults to ``(100, 100)``.
        angle:
            the angle of the stripe pattern in degrees. Defaults to ``45.0``.
        color1:
            the first color of the stripe pattern. Defaults to ``(0, 0, 0)``.
            the color can be specified as a tuple of ``(R, G, B)``,
            or a string (`e.g.,` ``"#ff0000"``, or ``"red"``).
        color2:
            the second color of the stripe pattern. Defaults to ``(255, 255, 255)``.
        opacity1:
            the opacity of the first color. Defaults to ``1.0``.
        opacity2:
            the opacity of the second color. Defaults to ``1.0``.
        total_width:
            the total width of the stripe pattern. The stripe pattern repeats every ``total_width``.
            Defaults to ``64.0``.
        phase:
            the phase of the stripe pattern in degrees. Defaults to ``0.0``.
        ratio:
            the ratio of the first color. Defaults to ``0.5``.
        duration:
            the duration of the layer. Defaults to ``1e6``.

    Animatable Attributes:
        ``angle``
        ``color1``
        ``color2``
        ``opacity1``
        ``opacity2``
        ``total_width``
        ``phase``
        ``ratio``
    """
    def __init__(
        self,
        size: tuple[int, int] = (100, 100),
        angle: float = 45.,
        color1: tuple[int, int, int] | str = (0, 0, 0),
        color2: tuple[int, int, int] | str = (255, 255, 255),
        opacity1: float = 1.0,
        opacity2: float = 1.0,
        total_width: float = 64.,
        phase: float = 0.,
        ratio: float = 0.5,
        duration: float = 1e6,
    ) -> None:
        self.size = size
        self.duration = duration
        self.angle = Attribute(angle, AttributeType.ANGLE)
        c1 = to_rgb(color1)
        c2 = to_rgb(color2)
        self.color1 = Attribute(c1, AttributeType.COLOR, range=(0., 255.))
        self.color2 = Attribute(c2, AttributeType.COLOR, range=(0., 255.))
        self.opacity1 = Attribute(opacity1, AttributeType.SCALAR, range=(0., 1.0))
        self.opacity2 = Attribute(opacity2, AttributeType.SCALAR, range=(0., 1.0))
        self.total_width = Attribute(total_width, AttributeType.SCALAR, range=(0., 1e6))
        self.phase = Attribute(phase, AttributeType.SCALAR)
        self.ratio = Attribute(ratio, AttributeType.SCALAR, range=(0., 1.0))

    def __call__(self, time: float) -> np.ndarray | None:
        if time < 0 or time >= self.duration:
            return None
        width, height = self.size
        ratio = float(self.ratio(time))
        c1 = np.concatenate([
            np.round(self.color1(time)).reshape(3, 1, 1),
            np.full((1, 1, 1), np.round(255 * float(self.opacity1(time))))], axis=0)
        c2 = np.concatenate([
            np.round(self.color2(time)).reshape(3, 1, 1),
            np.full((1, 1, 1), np.round(255 * float(self.opacity2(time))))], axis=0)
        if ratio <= 0.0:
            c1_img = np.broadcast_to(c1, (4, height, width)).transpose(1, 2, 0)
            return c1_img.astype(np.uint8)
        elif ratio >= 1.0:
            c2_img = np.broadcast_to(c2, (4, height, width)).transpose(1, 2, 0)
            return c2_img.astype(np.uint8)
        center = np.array([height / 2, width / 2])[:, None, None]
        inds = np.mgrid[:height, :width] - center
        theta = float(self.angle(time)) / 180.0 * np.pi
        phase = float(self.phase(time))
        stripe_width = float(self.total_width(time))

        eps = 1e-2
        v = np.array([np.sin(theta), np.cos(theta)], dtype=np.float64)
        p = (v.reshape(2, 1, 1) * inds).sum(axis=0, keepdims=True) / stripe_width + phase
        p = _fract(p)
        p = _smoothstep(max(ratio - eps, 0.0), min(ratio + eps, 1.0), p)
        color = p * np.broadcast_to(c1, (4, height, width)) + (1 - p) * np.broadcast_to(c2, (4, height, width))
        color = color.transpose(1, 2, 0).astype(np.uint8)
        return color


def _smoothstep(edge0: float, edge1: float, x: np.ndarray) -> np.ndarray:
    t = np.clip((x - edge0) / (edge1 - edge0), 0.0, 1.0)
    return t * t * (3.0 - 2.0 * t)


def _fract(x: np.ndarray) -> np.ndarray:
    return x - np.floor(x)



from __future__ import annotations

import hashlib
from os import PathLike
from pathlib import Path
from typing import Sequence

import numpy as np
from PIL import Image as PILImage

from ..imgproc import alpha_composite
from ..layer.mixin import TimelineMixin

try:
    from pdf2image import convert_from_path
    pdf2image_available = True
except ImportError:
    pdf2image_available = False


class Slide(TimelineMixin):
    """Slide layer for showing PDF slides.

    Many presentation videos use slides to explain the contents.
    This layer directly show the slides in the specified PDF file for convenience.

    .. note::
        This layer requires `pdf2image <https://github.com/Belval/pdf2image>`_ to be installed.

    Args:
        start_times:
            The start times for displaying slides.
        end_times:
            The end times for displaying slides.
        slide_file:
            The path to the PDF file containing the slides.
        slide_counter:
            The slide number for each slide. If ``None``, the slide number is automatically assigned.

    Examples:
        >>> from movis.contrib.commentary import Slide
        >>> # Show 0 page from 0.0 to 5.0 seconds, and 1 page from 5.0 to 10.0 seconds.
        >>> slide1 = Slide(start_times=[0.0, 5.0], end_times=[5.0, 10.0], slide_file="slides.pdf")
        >>> # Show 0 page from 0.0 to 5.0 seconds, and 1 page from 5.0 to 15.0 seconds.
        >>> slide2 = Slide([0., 5., 10.], [5., 10., 15.], "slides.pdf", slide_counter=[0, 1, 1])
    """
    def __init__(
        self,
        start_times: Sequence[float],
        end_times: Sequence[float],
        slide_file: str | PathLike,
        slide_counter: Sequence[int] | None = None,
    ) -> None:
        if not pdf2image_available:
            raise ImportError("pdf2image is not installed")
        super().__init__(start_times, end_times)
        if slide_counter is None:
            self.slide_timeline = np.arange(len(start_times))
        else:
            self.slide_timeline = np.asarray(slide_counter)
        self.slide_file = slide_file
        self.slide_images: list[np.ndarray] | None = None

    def get_key(self, time: float) -> int:
        """Return the key for caching."""
        idx = self.get_state(time)
        key = int(self.slide_timeline[idx]) if 0 < idx else -1
        return key

    def __call__(self, time: float) -> np.ndarray | None:
        idx = self.get_state(time)
        if idx < 0:
            return None
        slide_number = self.slide_timeline[idx]
        if self.slide_images is None:
            slide_images = []
            for img in convert_from_path(Path(self.slide_file)):
                img_np = np.asarray(img.convert("RGBA"))
                slide_images.append(img_np)
            self.slide_images = slide_images
        return self.slide_images[slide_number]


class Character(TimelineMixin):
    """Character layer for showing and animating characters.

    Some presentation videos may place characters who speak with different facial expressions
    depending on the situation. This layer mimics such a situation.

    Args:
        start_times:
            A list of start times in seconds at which the character's status will change.
            Should have the same length as `characters` and `character_status`.
        end_times:
            A list of end times in seconds corresponding to each status duration.
        character_name:
            The name of the character that this layer will manage.
        character_dir:
            The directory where images corresponding to the character's emotions are stored.
            For each emotion, there should be an image named `<emotion>.png`
            and an optional eye image named `<emotion>.eye.png`, `<emotion>.eye.0.png`, `<emotion>.eye.1.png`, ...
        characters:
            A list of character names for which the timeline events are defined.
            Should have the same length as ``start_times`` and ``character_status``.
        character_status:
            A list specifying the emotion or state of the character at the corresponding start time.
            Should have the same length as `start_times` and `characters`.
        initial_status:
            The initial emotion or state of the character before any timeline events occur.
            Defaults to "n" (Normal).
        blink_per_minute (int, optional):
            Number of times the character blinks per minute. Defaults to 3.
        blink_duration (float, optional):
            Duration of a single blink in seconds. Defaults to 0.2.

    Examples:
        >>> from movis.contrib.commentary import Character
        >>> # Show a character named "alice" from 0.0 to 10.0 seconds.
        >>> # Alice's emotion does not change and only displays "character/alice/n.png".
        >>> alice = Character([0.0, 5.0], [5.0, 10.0], "alice", "character/alice", ["alice", "bob"], ["n", "h"])
        >>> # Show a character named "bob" from 0.0 to 10.0 seconds.
        >>> # Initially Bob's emotion is "n" (Normal), but it changes to "h" (Happy) at 5.0 seconds.
        >>> # In this case, "n" means "character/bob/n.png" and "h" means "character/bob/h.png".
        >>> bob = Character([0.0, 5.0], [5.0, 10.0], "bob", "character/bob", ["alice", "bob"], ["n", "h"])
        >>> # If "character/bob" contains "n.eye.png" and "n.eye.0.png, "n.eye.1.png", ..., "n.eye.N.png",
        >>> # in the "n" state, Bob's eyes blink an average of three times per minute. and duration is 0.2 seconds.
    """
    def __init__(
        self,
        start_times: Sequence[float],
        end_times: Sequence[float],
        character_name: str,
        character_dir: str | PathLike,
        characters: Sequence[str],
        character_status: Sequence[str],
        initial_status: str = "n",
        blink_per_minute: int = 3,
        blink_duration: float = 0.2,
    ) -> None:
        assert len(start_times) == len(characters) == len(character_status)
        super().__init__(start_times, end_times)
        self.character_name = character_name
        self.character_imgs: dict[str, Path | np.ndarray] = {}
        self.eye_imgs: dict[str, list[Path | np.ndarray]] = {}
        character_dir = Path(character_dir)
        emotions = set()
        for character, status in zip(characters, character_status):
            if character == character_name:
                emotions.add(status)
        emotions.add(initial_status)
        for emotion in emotions:
            path = Path(character_dir) / f"{emotion}.png"
            self.character_imgs[emotion] = path
            eye_path = Path(character_dir) / f"{emotion}.eye.png"
            if eye_path.exists():
                eyes: list[Path | np.ndarray] = [eye_path]
                for f in character_dir.iterdir():
                    x = f.name.split(".")
                    if f.name.startswith(f"{emotion}.eye.") and len(x) == 4:
                        eyes.append(f)
                self.eye_imgs[emotion] = eyes

        self.character_timeline: list[str] = []
        status = initial_status
        for current_character, current_status in zip(characters, character_status):
            if current_character == character_name:
                status = current_status
            self.character_timeline.append(status)

        self.blink_per_minute = blink_per_minute
        self.blink_duration = blink_duration

    def _get_eye_state(self, time: float, idx: int) -> int:

        def rand_from_string(string: str, seed: int = 0) -> float:
            string = f"{seed}:{string}"
            s = hashlib.sha224(f"{seed}:{string}".encode("utf-8")).digest()
            x = np.frombuffer(s, dtype=np.uint32)[0]
            return np.random.RandomState(x).rand()

        emotion = self.character_timeline[idx]
        if emotion not in self.eye_imgs:
            return -1
        elif len(self.eye_imgs[emotion]) == 1:
            return 0
        p_threshold = self.blink_per_minute * self.blink_duration / 60
        n = int(time / self.blink_duration)
        p = rand_from_string(f"{self.character_name}:{n}")
        if p < p_threshold:
            frame_duration = self.blink_duration / (len(self.eye_imgs[emotion]) - 1)
            t1 = time - n * self.blink_duration
            n1 = int(t1 / frame_duration)
            return min(n1 + 1, len(self.eye_imgs[emotion]) - 1)
        else:
            return 0

    def get_key(self, time: float) -> tuple[str, int]:
        """Return the key for caching."""
        idx = self.get_state(time)
        if idx < 0:
            return ('', -1)
        emotion = self.character_timeline[idx]
        eye = self._get_eye_state(time, idx)
        return (emotion, eye)

    def __call__(self, time: float) -> np.ndarray | None:
        idx = self.get_state(time)
        if idx < 0:
            return None
        emotion = self.character_timeline[idx]
        character = self.character_imgs[emotion]
        if isinstance(character, Path):
            base_img = np.asarray(PILImage.open(character).convert("RGBA"))
            self.character_imgs[emotion] = base_img
        else:
            base_img = character

        if emotion in self.eye_imgs:
            eye_number = self._get_eye_state(time, idx)
            eye = self.eye_imgs[emotion][eye_number]
            if isinstance(eye, Path):
                eye_img = np.asarray(PILImage.open(eye).convert("RGBA"))
                self.eye_imgs[emotion][eye_number] = eye_img
            else:
                eye_img = eye
            base_img = alpha_composite(base_img, eye_img)
        return base_img



import numpy as np

import movis as mv


def test_concatenate():
    layer1 = mv.layer.Image.from_color(size=(10, 10), color=(255, 0, 0), duration=1.0)
    layer2 = mv.layer.Image.from_color(size=(10, 10), color=(0, 255, 0), duration=2.0)
    layer3 = mv.layer.Image.from_color(size=(10, 10), color=(0, 0, 255), duration=3.0)
    scene = mv.concatenate([layer1, layer2, layer3], size=(10, 10))
    assert scene.duration == 6.0
    assert scene.layers[0].duration == 1.0
    assert scene.layers[1].duration == 2.0
    assert scene.layers[2].duration == 3.0

    assert np.all(scene(0.0)[0, 0, :] == np.array([255, 0, 0, 255]))
    assert np.all(scene(1.0)[0, 0, :] == np.array([0, 255, 0, 255]))
    assert np.all(scene(3.0)[0, 0, :] == np.array([0, 0, 255, 255]))

    scene = mv.concatenate([layer1, layer2, layer3])
    assert scene.size == (10, 10)


def test_repeat():
    img1 = mv.layer.Image.from_color(size=(10, 10), color=(255, 0, 0), duration=1.0)
    img2 = mv.layer.Image.from_color(size=(10, 10), color=(0, 255, 0), duration=1.0)
    layer = mv.concatenate([img1, img2])
    scene = mv.repeat(layer, 3, size=(10, 10))

    assert scene.duration == 6.0
    assert scene.layers[0].duration == 2.0
    assert scene.layers[1].duration == 2.0
    assert scene.layers[2].duration == 2.0

    assert np.all(scene(0.0)[0, 0, :] == np.array([255, 0, 0, 255]))
    assert np.all(scene(1.0)[0, 0, :] == np.array([0, 255, 0, 255]))
    assert np.all(scene(2.0)[0, 0, :] == np.array([255, 0, 0, 255]))
    assert np.all(scene(3.0)[0, 0, :] == np.array([0, 255, 0, 255]))
    assert np.all(scene(4.0)[0, 0, :] == np.array([255, 0, 0, 255]))
    assert np.all(scene(5.0)[0, 0, :] == np.array([0, 255, 0, 255]))

    scene = mv.repeat(layer, 3)
    assert scene.size == (10, 10)


def test_trim():
    layer1 = mv.layer.Image.from_color(size=(10, 10), color=(255, 0, 0), duration=1.0)
    layer2 = mv.layer.Image.from_color(size=(10, 10), color=(0, 255, 0), duration=2.0)
    layer3 = mv.layer.Image.from_color(size=(10, 10), color=(0, 0, 255), duration=3.0)
    layer = mv.concatenate([layer1, layer2, layer3])
    scene = mv.trim(layer, start_times=[0.0, 3.0], end_times=[1.0, 4.0], size=(10, 10))

    assert scene.duration == 2.0
    assert len(scene) == 2
    assert scene.layers[0].duration == 1.0
    assert scene.layers[1].duration == 1.0

    assert np.all(scene(0.0)[0, 0, :] == np.array([255, 0, 0, 255]))
    assert np.all(scene(1.0)[0, 0, :] == np.array([0, 0, 255, 255]))

    scene = mv.trim(layer, start_times=[0.0, 3.0], end_times=[1.0, 4.0])
    assert scene.size == (10, 10)


def test_tile():
    layer1 = mv.layer.Image.from_color(size=(10, 10), color=(255, 0, 0), duration=1.0)
    layer2 = mv.layer.Image.from_color(size=(10, 10), color=(0, 255, 0), duration=2.0)
    scene = mv.tile([layer1, layer2], rows=1, cols=2)

    assert scene.duration == 2.0
    assert scene.size == (20, 10)
    assert len(scene) == 2
    assert scene.layers[0].duration == 1.0
    assert scene.layers[1].duration == 2.0

    assert np.all(scene(0.0)[0, 0, :] == np.array([255, 0, 0, 255]))
    assert np.all(scene(0.0)[0, 10, :] == np.array([0, 255, 0, 255]))

    scene = mv.tile([layer1, layer2], rows=2, cols=1)

    assert scene.duration == 2.0
    assert scene.size == (10, 20)
    assert len(scene) == 2
    assert scene.layers[0].duration == 1.0
    assert scene.layers[1].duration == 2.0

    assert np.all(scene(0.0)[0, 0, :] == np.array([255, 0, 0, 255]))
    assert np.all(scene(0.0)[10, 0, :] == np.array([0, 255, 0, 255]))


def test_crop():
    image = np.zeros((100, 100, 4), dtype=np.uint8)
    image[:, :, 3] = 255
    image[10:30, 10:30, 0] = 1
    layer = mv.layer.Image(image, duration=2.0)
    scene = mv.crop(layer, (10, 10, 20, 20))

    assert scene.duration == 2.0
    assert scene.size == (20, 20)
    assert len(scene) == 1

    frame = scene(0.0)
    assert np.all(frame[0, 0, :] == np.array([1, 0, 0, 255]))
    assert np.all(frame[-1, 0, :] == np.array([1, 0, 0, 255]))
    assert np.all(frame[0, -1, :] == np.array([1, 0, 0, 255]))
    assert np.all(frame[-1, -1, :] == np.array([1, 0, 0, 255]))


def test_switch():
    img1 = mv.layer.Image(np.full((100, 100, 4), 1, dtype=np.uint8), duration=5.0)
    img2 = mv.layer.Image(np.full((100, 100, 4), 2, dtype=np.uint8), duration=6.0)
    img3 = mv.layer.Image(np.full((100, 100, 4), 3, dtype=np.uint8), duration=7.0)
    scene = mv.switch([img1, img2, img3], [0.0, 1.0, 2.0, 3.0, 4.0], [0, 1, 2, 1, 0])

    assert scene.duration == 5.0
    assert scene.size == (100, 100)
    assert len(scene) == 5

    assert np.all(scene(0.0)[0, 0, :] == np.array([1, 1, 1, 1]))
    assert np.all(scene(1.0)[0, 0, :] == np.array([2, 2, 2, 2]))
    assert np.all(scene(2.0)[0, 0, :] == np.array([3, 3, 3, 3]))
    assert np.all(scene(3.0)[0, 0, :] == np.array([2, 2, 2, 2]))
    assert np.all(scene(4.0)[0, 0, :] == np.array([1, 1, 1, 1]))


def test_insert():
    img1 = mv.layer.Image(np.full((100, 100, 4), 1, dtype=np.uint8), duration=2.0)
    img2 = mv.layer.Image(np.full((100, 100, 4), 3, dtype=np.uint8), duration=3.0)
    source = mv.concatenate([img1, img2], size=img1.size)

    target = mv.layer.Image(np.full((100, 100, 4), 2, dtype=np.uint8), duration=1.0)
    scene = mv.insert(source, target, 2.0)

    assert scene.duration == 6.0
    assert scene.size == (100, 100)
    assert len(scene) == 3

    assert np.all(scene(0.0)[0, 0, :] == np.array([1, 1, 1, 1]))
    assert np.all(scene(2.0)[0, 0, :] == np.array([2, 2, 2, 2]))
    assert np.all(scene(3.0)[0, 0, :] == np.array([3, 3, 3, 3]))
    assert np.all(scene(6.0 - 1e-5)[0, 0, :] == np.array([3, 3, 3, 3]))



from __future__ import annotations

import difflib
import hashlib
from os import PathLike
from pathlib import Path
from typing import Hashable

import librosa

try:
    from pandas import DataFrame
    pandas_available = True
except ImportError:
    pandas_available = False


def make_voicevox_dataframe(audio_dir: str | PathLike) -> "DataFrame":
    """Create a ``pandas.DataFrame`` representing the timeline of audio files generated by Voicevox.

    This function takes a directory containing `.wav` files generated by `Voicevox <https://voicevox.hiroshiba.jp/>`_,
    and returns a DataFrame where each row represents an audio file, sorted by the filename.
    The ``DataFrame`` will have columns ``start_time``, ``end_time``, and ``audio_file``.

    Args:
        audio_dir:
            The directory containing `.wav` files generated by Voicevox.

    Returns:
        A DataFrame with columns `start_time`, `end_time`, and `audio_file`.

        - ``start_time``: The start time of the audio clip in seconds.
        - ``end_time``: The end time of the audio clip in seconds.
        - ``audio_file``: The path to the audio file.

    Example:
        >>> from movis.contrib.voicevox import make_voicevox_dataframe
        >>> make_voicevox_dataframe('audio')
        start_time   end_time                   audio_file
        0    0.000000   4.682667  audio/001_ずんだもん（ノーマル）.wav
        1    4.682667   8.202667    audio/002_四国めたん（ノーマル）.wav
        2    8.202667  13.930667  audio/003_ずんだもん（ノーマル）.wav
        3   13.930667  18.133333    audio/004_四国めたん（ノーマル）.wav
        4   18.133333  20.138667    audio/005_四国めたん（ノーマル）.wav
        5   20.138667  25.205333  audio/007_ずんだもん（ノーマル）.wav
        6   25.205333  30.410667  audio/008_ずんだもん（ノーマル）.wav
    """
    if not pandas_available:
        raise ImportError("pandas is not installed")

    def get_audio_length(filename: str | PathLike) -> float:
        return librosa.get_duration(path=filename)

    wav_files = sorted(f for f in Path(audio_dir).iterdir() if f.suffix == ".wav")
    rows = []
    start_time = 0.0
    for wav_file in wav_files:
        duration = get_audio_length(wav_file)
        end_time = start_time + duration
        dic = {
            "start_time": start_time,
            "end_time": end_time,
        }
        rows.append(dic)
        start_time = end_time
    frame = DataFrame(rows)
    frame["audio_file"] = [str(p) for p in wav_files]
    return frame


def make_timeline_from_voicevox(
    audio_dir: str | PathLike,
    max_text_length: int = 25,
    extra_columns: tuple[tuple[str, Hashable], ...] = (
        ("slide", 0), ("status", "n")),
) -> "DataFrame":
    """Create a ``pandas.DataFrame`` based on text files generated by `Voicevox <https://voicevox.hiroshiba.jp/>`_.

    This function reads ``.txt`` files from a directory generated by Voicevox and constructs a ``DataFrame``.
    Each row of the ``DataFrame`` corresponds to a text file and includes columns such as ``'character'``,
    ``'hash'``, and ``'text'``, as well as any additional columns specified in ``extra_columns``.

    Args:
        audio_dir:
            The directory containing ``.txt`` files generated by Voicevox.
        max_text_length:
            The maximum length of text for each entry in the ``'text'`` column.
            Strings longer than ``max_text_length`` are automatically line-breaked
            for subtitling convenience. Defaults to 25.
        extra_columns:
            Additional columns to be added to the DataFrame. Each inner tuple should contain a
            column name and a default value. Defaults to ``(("slide", 0), ("status", "n"))``.
            For example, ``extra_columns=(("slide", 0), ("status", "n"))`` will add two columns
            named ``'slide'`` and ``'status'`` with default values ``0`` and ``'n'``, respectively.

    .. note::
        The ``'character'`` column is automatically added to the DataFrame based on the filename.
        For example, if the filename is ``'001_ずんだもん（ノーマル）.txt'``, the ``'character'`` column
        will be set to ``'zunda'``.

    .. note::
        The ``'hash'`` column is automatically added to the DataFrame based on the content of the text file.
        This is used in ``merget_timeline`` to detect changes in the text file.

    Returns:
        A DataFrame with columns that may include 'character', 'hash', 'text', and any additional columns.

        - ``character``: The identifier for the Voicevox character.
        - ``hash``: A hash prefix based on the content of the text file.
        - ``text``: The text content, potentially divided into multiple parts if it exceeds ``max_text_length``.
        - Any additional columns specified in ``extra_columns``.

    Examples:
        >>> from movis.contrib.voicevox import make_timeline_from_voicevox
        >>> make_timeline_from_voicevox('audio')
        character    hash                                      text  slide status
        0     zunda  9bc5d6         こんにちは！このサンプルでは、簡単なずんだもん動画\\nを作ります。      0      n
        1     metan  ca647c                  このような形で対談を行うのが解説動画の特徴です。      0      n
        2     zunda  75d695    先生役と生徒役で対談することで、内容をわかりやすく\\n視聴者に伝えられます。      0      n
        3     metan  4d880b          また疑問を生徒側が投げかけることで、より深い解説を\\nスムーズに      0      n
        4     metan  e9e4f0                        伝えられるのも利点と言えるでしょう。      0      n
        5     zunda  702873  欠点としては、現状のところ深い内容を解説する動画に\\nは作られていないことです。      0      n
        6     zunda  4a1a86   手間がかかることが理由の一つかもしれませんが、この\\nようなライブラリを通して      0      n
        7     zunda  cb26fc     より多くのユーザー層へ、専門的な知識を伝えられるこ\\nとを期待しています。      0      n
    """
    if not pandas_available:
        raise ImportError("pandas is not installed")

    def get_paths(src_dir: Path, ext: str) -> list[Path]:
        src_dir = Path(src_dir)
        return sorted(f for f in src_dir.iterdir() if f.suffix == ext)

    def get_hash_prefix(text):
        text_bytes = text.encode("utf-8")
        sha1_hash = hashlib.sha1(text_bytes)
        hashed_text = sha1_hash.hexdigest()
        prefix = hashed_text[:6]
        return prefix

    txt_files = get_paths(Path(audio_dir), ".txt")
    lines = []
    for txt_file in txt_files:
        raw_text = open(txt_file, "r", encoding="utf-8-sig").read()
        if raw_text == "":
            raise RuntimeError(
                f"Empty text file: {txt_file}. Please remove it and try again."
            )
        character_dict = {
            "ずんだもん": "zunda",
            "四国めたん": "metan",
            "春日部つむぎ": "tsumugi",
        }
        character = txt_file.stem.split("_")[1].split("（")[0]
        text = "\\n".join([
            raw_text[i: i + max_text_length]
            for i in range(0, len(raw_text), max_text_length)]
        )
        dic = {
            "character": character_dict[character],
            "hash": get_hash_prefix(raw_text),
            "text": text,
        }
        for column_name, default_value in extra_columns:
            dic[column_name] = default_value
        lines.append(dic)
    return DataFrame(lines)


def merge_timeline(
    old_timeline: "DataFrame",
    new_timeline: "DataFrame",
    key="hash",
    description="text",
) -> "DataFrame":
    """Merge an old DataFrame with a new DataFrame to update its timeline information.

    This function takes an old DataFrame and a new DataFrame, both generated by `make_timeline_from_voicevox`,
    and merges them to create an updated DataFrame. Rows in both DataFrames are compared based on the
    specified key column. Any discrepancies between the old and new DataFrames are flagged in the
    specified description column with ``">>>>>"`` for added rows and ``"<<<<<"`` for removed rows.

    Args:
        old_timeline:
            The old DataFrame to be updated.
        new_timeline:
            The new DataFrame to update the old one.
        key:
            The column used for comparison. Defaults to "hash".
        description:
            The column where discrepancies will be flagged. Defaults to "text".

    Returns:
        An updated DataFrame that merges the old and new timelines.

    Example:
        >>> import pandas as pd
        >>> from movis.contrib.voicevox import merge_timeline
        >>> old_timeline = pd.DataFrame({'hash': [1, 2], 'text': ['a', 'b']})
        >>> new_timeline = pd.DataFrame({'hash': [2, 3], 'text': ['b', 'c']})
        >>> merge_timeline(old_timeline, new_timeline)
        hash     text
        0     1  <<<<< a
        1     2        b
        1     3  >>>>> c
    """
    if not pandas_available:
        raise ImportError("pandas is not installed")

    differ = difflib.Differ()
    diff = differ.compare(old_timeline[key].to_list(), new_timeline[key].tolist())
    result = []
    old_indices = old_timeline.index.tolist()
    new_indices = new_timeline.index.tolist()
    old_idx, new_idx = 0, 0
    for d in diff:
        if d.startswith("-"):
            row = old_timeline.iloc[old_indices[old_idx]].copy()
            row[description] = f"<<<<< {row[description]}"
            result.append(row)
            old_idx += 1
        elif d.startswith("+"):
            row = new_timeline.iloc[new_indices[new_idx]].copy()
            row[description] = f">>>>> {row[description]}"
            result.append(row)
            new_idx += 1
        else:
            result.append(old_timeline.iloc[old_indices[old_idx]])
            old_idx += 1
            new_idx += 1
    return DataFrame(result)



import numpy as np
import pytest

from movis.attribute import Attribute
from movis.transform import (Direction, Transform, TransformValue,
                             transform_to_1dscalar, transform_to_2dvector,
                             transform_to_3dvector)


def test_anchor_point():
    transform = Transform(anchor_point=(0.5, 0.75))
    assert np.all(transform.anchor_point.init_value == (0.5, 0.75))
    assert np.all(transform.anchor_point(0.0) == (0.5, 0.75))
    assert np.all(transform.anchor_point(0.5) == (0.5, 0.75))
    assert np.all(transform.anchor_point(1.0) == (0.5, 0.75))


def test_position():
    transform = Transform(position=(0.5, 1.0))
    assert np.all(transform.position.init_value == (0.5, 1.0))
    assert np.all(transform.position(0.0) == (0.5, 1.0))
    assert np.all(transform.position(0.5) == (0.5, 1.0))
    assert np.all(transform.position(1.0) == (0.5, 1.0))


def test_scale():
    transform = Transform(scale=0.5)
    assert np.all(transform.scale.init_value == (0.5, 0.5))
    assert np.all(transform.scale(0.0) == (0.5, 0.5))
    assert np.all(transform.scale(0.5) == (0.5, 0.5))
    assert np.all(transform.scale(1.0) == (0.5, 0.5))

    transform = Transform(scale=(0.5, 1.0))
    assert np.all(transform.scale.init_value == (0.5, 1.0))
    assert np.all(transform.scale(0.0) == (0.5, 1.0))
    assert np.all(transform.scale(0.5) == (0.5, 1.0))
    assert np.all(transform.scale(1.0) == (0.5, 1.0))


def test_rotation():
    transform = Transform(rotation=45.0)
    assert np.all(transform.rotation.init_value == 45.0)
    assert np.all(transform.rotation(0.0) == 45.0)
    assert np.all(transform.rotation(0.5) == 45.0)
    assert np.all(transform.rotation(1.0) == 45.0)


def test_opacity():
    transform = Transform(opacity=0.5)
    assert np.all(transform.opacity.init_value == 0.5)
    assert np.all(transform.opacity(0.0) == 0.5)
    assert np.all(transform.opacity(0.5) == 0.5)
    assert np.all(transform.opacity(1.0) == 0.5)


def test_attributes():
    transform = Transform()
    assert isinstance(transform.attributes, dict)
    assert len(transform.attributes) == 5
    assert "anchor_point" in transform.attributes
    assert "position" in transform.attributes
    assert "scale" in transform.attributes
    assert "rotation" in transform.attributes
    assert "opacity" in transform.attributes


def test_get_current_value():
    transform = Transform()
    assert isinstance(transform.get_current_value(0.0), TransformValue)
    assert isinstance(transform.get_current_value(0.5), TransformValue)
    assert isinstance(transform.get_current_value(1.0), TransformValue)

    v = transform.get_current_value(0.0)
    assert isinstance(v.anchor_point, tuple)
    assert isinstance(v.position, tuple)
    assert isinstance(v.scale, tuple)
    assert isinstance(v.rotation, float)
    assert isinstance(v.opacity, float)
    assert isinstance(v.origin_point, Direction)


def test_get_attributes():
    transform = Transform()
    attrs = transform.attributes
    assert isinstance(attrs, dict)
    assert len(attrs) == 5

    for attr in attrs.values():
        assert isinstance(attr, Attribute)


def test_from_positions():
    transform = Transform.from_positions((100, 100))
    assert isinstance(transform, Transform)
    assert np.all(transform.position.init_value == np.array([50.0, 50.0]))

    transform = Transform.from_positions((100, 100), top=0.0)
    assert isinstance(transform, Transform)
    assert np.all(transform.position.init_value == np.array([50.0, 0.0]))
    assert transform.origin_point == Direction.TOP_CENTER

    transform = Transform.from_positions((100, 100), bottom=0.0)
    assert isinstance(transform, Transform)
    assert np.all(transform.position.init_value == np.array([50.0, 100.0]))
    assert transform.origin_point == Direction.BOTTOM_CENTER

    transform = Transform.from_positions((100, 100), left=0.0)
    assert isinstance(transform, Transform)
    assert np.all(transform.position.init_value == np.array([0.0, 50.0]))
    assert transform.origin_point == Direction.CENTER_LEFT

    transform = Transform.from_positions((100, 100), right=0.0)
    assert isinstance(transform, Transform)
    assert np.all(transform.position.init_value == np.array([100.0, 50.0]))
    assert transform.origin_point == Direction.CENTER_RIGHT

    transform = Transform.from_positions((100, 100), top=0.0, left=0.0)
    assert isinstance(transform, Transform)
    assert np.all(transform.position.init_value == np.array([0.0, 0.0]))
    assert transform.origin_point == Direction.TOP_LEFT

    transform = Transform.from_positions((100, 100), top=0.0, right=0.0)
    assert isinstance(transform, Transform)
    assert np.all(transform.position.init_value == np.array([100.0, 0.0]))
    assert transform.origin_point == Direction.TOP_RIGHT

    transform = Transform.from_positions((100, 100), bottom=0.0, left=0.0)
    assert isinstance(transform, Transform)
    assert np.all(transform.position.init_value == np.array([0.0, 100.0]))
    assert transform.origin_point == Direction.BOTTOM_LEFT

    transform = Transform.from_positions((100, 100), bottom=0.0, right=0.0)
    assert isinstance(transform, Transform)
    assert np.all(transform.position.init_value == np.array([100.0, 100.0]))
    assert transform.origin_point == Direction.BOTTOM_RIGHT


def test_transform_value_defaults():
    transform = TransformValue()
    assert transform.anchor_point == (0.0, 0.0)
    assert transform.position == (0.0, 0.0)
    assert transform.scale == (1.0, 1.0)
    assert transform.rotation == 0.0
    assert transform.opacity == 1.0
    assert transform.origin_point == Direction.CENTER


def test_transform_value_custom_values():
    transform = TransformValue(anchor_point=(1.0, 2.0), position=(3.0, 4.0), scale=(2.0, 2.0),
                               rotation=45.0, opacity=0.5, origin_point=Direction.TOP_LEFT)
    assert transform.anchor_point == (1.0, 2.0)
    assert transform.position == (3.0, 4.0)
    assert transform.scale == (2.0, 2.0)
    assert transform.rotation == 45.0
    assert transform.opacity == 0.5
    assert transform.origin_point == Direction.TOP_LEFT


def test_transform_to_1dscalar():
    assert transform_to_1dscalar(3.0) == 3.0
    assert transform_to_1dscalar([3.0]) == 3.0
    assert transform_to_1dscalar((3.0,)) == 3.0
    assert transform_to_1dscalar(np.array(3.0)) == 3.0
    assert transform_to_1dscalar(np.array([3.0])) == 3.0

    with pytest.raises(ValueError):
        transform_to_1dscalar([])


def test_transform_to_2dvector():
    assert transform_to_2dvector(3.0) == (3.0, 3.0)
    assert transform_to_2dvector([3.0]) == (3.0, 3.0)
    assert transform_to_2dvector((3.0,)) == (3.0, 3.0)
    assert transform_to_2dvector([3.0, 4.0]) == (3.0, 4.0)
    assert transform_to_2dvector(np.array(3.0)) == (3.0, 3.0)
    assert transform_to_2dvector(np.array([3.0])) == (3.0, 3.0)

    with pytest.raises(ValueError):
        transform_to_2dvector([3.0, 4.0, 5.0])
    with pytest.raises(ValueError):
        transform_to_2dvector([])


def test_transform_to_3dvector():
    assert transform_to_3dvector(3.0) == (3.0, 3.0, 3.0)
    assert transform_to_3dvector([3.0]) == (3.0, 3.0, 3.0)
    assert transform_to_3dvector((3.0,)) == (3.0, 3.0, 3.0)
    assert transform_to_3dvector([3.0, 4.0, 5.0]) == (3.0, 4.0, 5.0)
    assert transform_to_3dvector(np.array(3.0)) == (3.0, 3.0, 3.0)
    assert transform_to_3dvector(np.array([3.0])) == (3.0, 3.0, 3.0)

    with pytest.raises(ValueError):
        transform_to_3dvector([3.0, 4.0])
    with pytest.raises(ValueError):
        transform_to_3dvector([3.0, 4.0, 5.0, 6.0])
    with pytest.raises(ValueError):
        transform_to_3dvector([])
